# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hypothesis import HealthCheck, given, settings
import pytest

from swh.model.hypothesis_strategies import branch_names, snapshot_targets, snapshots
from swh.model.model import (
    OriginVisit,
    OriginVisitStatus,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
)
from swh.storage.algos.snapshot import (
    snapshot_get_all_branches,
    snapshot_get_latest,
    snapshot_id_get_from_revision,
    snapshot_resolve_alias,
    visits_and_snapshots_get_from_revision,
)
from swh.storage.utils import now

# list of hypothesis disabled health checks in some of the hypthesis driven tests here
disabled_health_checks = []
# we use getattr here to keep mypy happy regardless hypothesis version
if hasattr(HealthCheck, "function_scoped_fixture"):
    disabled_health_checks.append(HealthCheck.function_scoped_fixture)


@pytest.fixture
def swh_storage_backend_config():
    yield {
        "cls": "memory",
        "journal_writer": None,
    }


@settings(suppress_health_check=disabled_health_checks)
@given(snapshot=snapshots(min_size=0, max_size=10, only_objects=False))
def test_snapshot_small(swh_storage, snapshot):  # noqa
    swh_storage.snapshot_add([snapshot])

    returned_snapshot = snapshot_get_all_branches(swh_storage, snapshot.id)
    assert snapshot == returned_snapshot


@settings(suppress_health_check=disabled_health_checks, deadline=None)
@given(branch_name=branch_names(), snapshot_target=snapshot_targets(only_objects=True))
def test_snapshot_large(swh_storage, branch_name, snapshot_target):  # noqa
    snapshot = Snapshot(
        branches={b"%s%05d" % (branch_name, i): snapshot_target for i in range(10000)},
    )

    swh_storage.snapshot_add([snapshot])

    returned_snapshot = snapshot_get_all_branches(swh_storage, snapshot.id)
    assert snapshot == returned_snapshot


def test_snapshot_get_latest_none(swh_storage, sample_data):
    """Retrieve latest snapshot on unknown origin or origin without snapshot should
    yield no result

    """
    # unknown origin so None
    assert snapshot_get_latest(swh_storage, "unknown-origin") is None

    # no snapshot on origin visit so None
    origin = sample_data.origin
    swh_storage.origin_add(sample_data.origins)
    origin_visit, origin_visit2 = sample_data.origin_visits[:2]
    assert origin_visit.origin == origin.url

    swh_storage.origin_visit_add(sample_data.origin_visits)
    assert snapshot_get_latest(swh_storage, origin.url) is None

    # no visit status for now (origin visits in sample_data do have a visit_id,
    # so origin_visit_add won't auto-add the OriginVisitStatus:
    assert swh_storage.origin_visit_get_latest(origin.url) is None

    # add "created" visit statuses
    swh_storage.origin_visit_status_add(sample_data.origin_visit_statuses)
    # now we should have a result from origin_visit_get_latest()
    ov1 = swh_storage.origin_visit_get_latest(origin.url)
    assert ov1 is not None

    # visit references a snapshot but the snapshot does not exist in backend for some
    # reason
    complete_snapshot = sample_data.snapshots[2]
    swh_storage.origin_visit_status_add(
        [
            OriginVisitStatus(
                origin=origin.url,
                visit=ov1.visit,
                date=origin_visit2.date,
                status="partial",
                snapshot=complete_snapshot.id,
            )
        ]
    )
    # so we do not find it
    assert snapshot_get_latest(swh_storage, origin.url) is None
    assert snapshot_get_latest(swh_storage, origin.url, branches_count=1) is None


def test_snapshot_get_latest(swh_storage, sample_data):
    origin = sample_data.origin
    swh_storage.origin_add(sample_data.origins)

    visit1, visit2 = sample_data.origin_visits[:2]
    assert visit1.origin == origin.url

    swh_storage.origin_visit_add([visit1])
    swh_storage.origin_visit_status_add([sample_data.origin_visit_status])
    ov1 = swh_storage.origin_visit_get_latest(origin.url)

    # Add snapshot to visit1, latest snapshot = visit 1 snapshot
    complete_snapshot = sample_data.snapshots[2]
    swh_storage.snapshot_add([complete_snapshot])

    swh_storage.origin_visit_status_add(
        [
            OriginVisitStatus(
                origin=origin.url,
                visit=ov1.visit,
                date=visit2.date,
                status="partial",
                snapshot=None,
            )
        ]
    )
    assert visit1.date < visit2.date

    # no snapshot associated to the visit, so None
    actual_snapshot = snapshot_get_latest(
        swh_storage, origin.url, allowed_statuses=["partial"]
    )
    assert actual_snapshot is None

    date_now = now()
    assert visit2.date < date_now
    swh_storage.origin_visit_status_add(
        [
            OriginVisitStatus(
                origin=ov1.origin,
                visit=ov1.visit,
                date=date_now,
                type=ov1.type,
                status="full",
                snapshot=complete_snapshot.id,
            )
        ]
    )

    swh_storage.origin_visit_add(
        [
            OriginVisit(
                origin=origin.url,
                date=now(),
                type=visit1.type,
            )
        ]
    )

    actual_snapshot = snapshot_get_latest(swh_storage, origin.url)
    assert actual_snapshot is not None
    assert actual_snapshot == complete_snapshot

    actual_snapshot = snapshot_get_latest(swh_storage, origin.url, branches_count=1)
    assert actual_snapshot is not None
    assert actual_snapshot.id == complete_snapshot.id
    assert len(actual_snapshot.branches.values()) == 1

    with pytest.raises(ValueError, match="branches_count must be a positive integer"):
        snapshot_get_latest(swh_storage, origin.url, branches_count="something-wrong")


def test_snapshot_get_latest_for_visit_type(swh_storage, sample_data):
    origin = sample_data.origin
    swh_storage.origin_add(sample_data.origins)

    visit1, visit2 = sample_data.origin_visits[:2]
    visit1 = visit1.evolve(type="git")
    visit2 = visit2.evolve(type="git-checkout")
    assert visit1.origin == visit2.origin == origin.url

    ov1, ov2 = swh_storage.origin_visit_add([visit1, visit2])

    simple_snapshot = sample_data.snapshots[0]
    complete_snapshot = sample_data.snapshots[2]
    swh_storage.snapshot_add([simple_snapshot, complete_snapshot])

    swh_storage.origin_visit_status_add(
        [
            OriginVisitStatus(
                origin=origin.url,
                visit=ov1.visit,
                date=visit1.date,
                status="full",
                snapshot=complete_snapshot.id,
            )
        ]
    )

    swh_storage.origin_visit_status_add(
        [
            OriginVisitStatus(
                origin=origin.url,
                visit=ov2.visit,
                date=visit2.date,
                status="full",
                snapshot=simple_snapshot.id,
            )
        ]
    )

    git_snapshot = snapshot_get_latest(
        swh_storage,
        origin.url,
        visit_type="git",
    )
    assert git_snapshot == complete_snapshot

    git_checkout_snapshot = snapshot_get_latest(
        swh_storage,
        origin.url,
        visit_type="git-checkout",
    )
    assert git_checkout_snapshot == simple_snapshot


def test_snapshot_id_get_from_revision(swh_storage, sample_data):
    origin = sample_data.origin
    swh_storage.origin_add([origin])

    date_visit2 = now()
    visit1, visit2 = sample_data.origin_visits[:2]
    assert visit1.origin == origin.url

    ov1, ov2 = swh_storage.origin_visit_add([visit1, visit2])

    revision1, revision2, revision3 = sample_data.revisions[:3]
    swh_storage.revision_add([revision1, revision2])

    empty_snapshot, complete_snapshot = sample_data.snapshots[1:3]
    swh_storage.snapshot_add([complete_snapshot])

    # Add complete_snapshot to visit1 which targets revision1
    ovs1, ovs2 = [
        OriginVisitStatus(
            origin=ov1.origin,
            visit=ov1.visit,
            date=date_visit2,
            type=ov1.type,
            status="partial",
            snapshot=complete_snapshot.id,
        ),
        OriginVisitStatus(
            origin=ov2.origin,
            visit=ov2.visit,
            date=now(),
            type=ov2.type,
            status="full",
            snapshot=empty_snapshot.id,
        ),
    ]

    swh_storage.origin_visit_status_add([ovs1, ovs2])
    assert ov1.date < ov2.date
    assert ov2.date < ovs1.date
    assert ovs1.date < ovs2.date

    # revision3 does not exist so result is None
    actual_snapshot_id = snapshot_id_get_from_revision(
        swh_storage, origin.url, revision3.id
    )
    assert actual_snapshot_id is None

    # no snapshot targets revision2 for origin.url so result is None
    actual_snapshot_id = snapshot_id_get_from_revision(
        swh_storage, origin.url, revision2.id
    )
    assert actual_snapshot_id is None

    # complete_snapshot targets at least revision1
    actual_snapshot_id = snapshot_id_get_from_revision(
        swh_storage, origin.url, revision1.id
    )
    assert actual_snapshot_id == complete_snapshot.id


def test_visit_and_snapshot_get_from_revision(swh_storage, sample_data):
    origin = sample_data.origin
    swh_storage.origin_add([origin])

    date_visit2 = now()
    visit1, visit2 = sample_data.origin_visits[:2]
    assert visit1.origin == origin.url

    ov1, ov2 = swh_storage.origin_visit_add([visit1, visit2])

    revision1, revision2, revision3 = sample_data.revisions[:3]
    swh_storage.revision_add([revision1, revision2])

    empty_snapshot, complete_snapshot = sample_data.snapshots[1:3]
    swh_storage.snapshot_add([complete_snapshot])

    # Add complete_snapshot to visit1 which targets revision1
    ovs1, ovs2 = [
        OriginVisitStatus(
            origin=ov1.origin,
            visit=ov1.visit,
            date=date_visit2,
            type=ov1.type,
            status="partial",
            snapshot=complete_snapshot.id,
        ),
        OriginVisitStatus(
            origin=ov2.origin,
            visit=ov2.visit,
            date=now(),
            type=ov2.type,
            status="full",
            snapshot=empty_snapshot.id,
        ),
    ]

    swh_storage.origin_visit_status_add([ovs1, ovs2])
    assert ov1.date < ov2.date
    assert ov2.date < ovs1.date
    assert ovs1.date < ovs2.date

    # revision3 does not exist so result is None
    actual_snapshot_id = snapshot_id_get_from_revision(
        swh_storage, origin.url, revision3.id
    )
    assert actual_snapshot_id is None

    # no snapshot targets revision2 for origin.url so result is None
    res = list(
        visits_and_snapshots_get_from_revision(swh_storage, origin.url, revision2.id)
    )
    assert res == []

    # complete_snapshot targets at least revision1
    res = list(
        visits_and_snapshots_get_from_revision(swh_storage, origin.url, revision1.id)
    )
    assert res == [(ov1, ovs1, complete_snapshot)]


def test_snapshot_resolve_aliases_unknown_snapshot(swh_storage):
    assert snapshot_resolve_alias(swh_storage, b"foo", b"HEAD") is None


def test_snapshot_resolve_aliases_no_aliases(swh_storage):
    snapshot = Snapshot(branches={})
    swh_storage.snapshot_add([snapshot])

    assert snapshot_resolve_alias(swh_storage, snapshot.id, b"HEAD") is None


def test_snapshot_resolve_alias(swh_storage, sample_data):
    rev_branch_name = b"revision_branch"
    rel_branch_name = b"release_branch"
    rev_alias1_name = b"rev_alias1"
    rev_alias2_name = b"rev_alias2"
    rev_alias3_name = b"rev_alias3"
    rel_alias_name = b"rel_alias"
    rev_branch_info = SnapshotBranch(
        target=sample_data.revisions[0].id,
        target_type=SnapshotTargetType.REVISION,
    )
    rel_branch_info = SnapshotBranch(
        target=sample_data.releases[0].id,
        target_type=SnapshotTargetType.RELEASE,
    )
    rev_alias1_branch_info = SnapshotBranch(
        target=rev_branch_name, target_type=SnapshotTargetType.ALIAS
    )
    rev_alias2_branch_info = SnapshotBranch(
        target=rev_alias1_name, target_type=SnapshotTargetType.ALIAS
    )

    rev_alias3_branch_info = SnapshotBranch(
        target=rev_alias2_name, target_type=SnapshotTargetType.ALIAS
    )
    rel_alias_branch_info = SnapshotBranch(
        target=rel_branch_name, target_type=SnapshotTargetType.ALIAS
    )

    snapshot = Snapshot(
        branches={
            rev_branch_name: rev_branch_info,
            rel_branch_name: rel_branch_info,
            rev_alias1_name: rev_alias1_branch_info,
            rev_alias2_name: rev_alias2_branch_info,
            rev_alias3_name: rev_alias3_branch_info,
            rel_alias_name: rel_alias_branch_info,
        }
    )
    swh_storage.snapshot_add([snapshot])

    for alias_name, expected_branch in (
        (rev_alias1_name, rev_branch_info),
        (
            rev_alias2_name,
            rev_branch_info,
        ),
        (
            rev_alias3_name,
            rev_branch_info,
        ),
        (rel_alias_name, rel_branch_info),
    ):
        assert (
            snapshot_resolve_alias(swh_storage, snapshot.id, alias_name)
            == expected_branch
        )


def test_snapshot_resolve_alias_dangling_branch(swh_storage):
    dangling_branch_name = b"dangling_branch"
    alias_name = b"rev_alias"

    alias_branch = SnapshotBranch(
        target=dangling_branch_name, target_type=SnapshotTargetType.ALIAS
    )

    snapshot = Snapshot(
        branches={
            dangling_branch_name: None,
            alias_name: alias_branch,
        }
    )
    swh_storage.snapshot_add([snapshot])

    assert snapshot_resolve_alias(swh_storage, snapshot.id, alias_name) is None


def test_snapshot_resolve_alias_missing_branch(swh_storage):
    missing_branch_name = b"missing_branch"
    alias_name = b"rev_alias"

    alias_branch = SnapshotBranch(
        target=missing_branch_name, target_type=SnapshotTargetType.ALIAS
    )

    snapshot = Snapshot(
        id=b"42" * 10,
        branches={
            alias_name: alias_branch,
        },
    )
    swh_storage.snapshot_add([snapshot])

    assert snapshot_resolve_alias(swh_storage, snapshot.id, alias_name) is None


def test_snapshot_resolve_alias_cycle_found(swh_storage):
    alias1_name = b"alias_1"
    alias2_name = b"alias_2"
    alias3_name = b"alias_3"
    alias4_name = b"alias_4"

    alias1_branch_info = SnapshotBranch(
        target=alias2_name, target_type=SnapshotTargetType.ALIAS
    )
    alias2_branch_info = SnapshotBranch(
        target=alias3_name, target_type=SnapshotTargetType.ALIAS
    )
    alias3_branch_info = SnapshotBranch(
        target=alias4_name, target_type=SnapshotTargetType.ALIAS
    )
    alias4_branch_info = SnapshotBranch(
        target=alias2_name, target_type=SnapshotTargetType.ALIAS
    )

    snapshot = Snapshot(
        branches={
            alias1_name: alias1_branch_info,
            alias2_name: alias2_branch_info,
            alias3_name: alias3_branch_info,
            alias4_name: alias4_branch_info,
        }
    )
    swh_storage.snapshot_add([snapshot])

    assert snapshot_resolve_alias(swh_storage, snapshot.id, alias1_name) is None



# Copyright (C) 2024-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Tests the 'table_options' configuration item from the config is applied
when creating tables"""

import time

import pytest

from swh.model.swhids import ExtendedSWHID
from swh.storage.interface import ObjectReference


@pytest.fixture
def swh_storage_backend_config(swh_storage_cassandra_backend_config):
    return {
        **swh_storage_cassandra_backend_config,
        "table_options": {
            "release": "default_time_to_live = 2",
            "object_references_*": "default_time_to_live = 2",
        },
    }


@pytest.fixture
def swh_storage(swh_storage_backend):
    # need to drop and recreate object references partition for the
    # defined TTLs to be applied to cassandra tables
    for partition in swh_storage_backend.object_references_list_partitions():
        swh_storage_backend.object_references_drop_partition(partition)
        swh_storage_backend.object_references_create_partition(
            partition.year, partition.week
        )
    yield swh_storage_backend
    for partition in swh_storage_backend.object_references_list_partitions():
        swh_storage_backend.object_references_drop_partition(partition)


@pytest.mark.cassandra
class TestCassandraStorageTtl:
    def test_release_ttl(self, swh_storage, sample_data):
        release, release2 = sample_data.releases[:2]

        init_missing = swh_storage.release_missing([release.id, release2.id])
        assert list(init_missing) == [release.id, release2.id]

        swh_storage.release_add([release])

        time.sleep(3)

        swh_storage.release_add([release2])

        end_missing = swh_storage.release_missing([release.id, release2.id])
        assert list(end_missing) == []

        time.sleep(3)

        end_missing = swh_storage.release_missing([release.id])
        assert list(end_missing) == []

    def test_object_references_ttl(self, swh_storage):
        source = ExtendedSWHID.from_string(f"swh:1:dir:{0:040x}")
        target = ExtendedSWHID.from_string(f"swh:1:cnt:{1:040x}")

        assert swh_storage.object_references_add(
            [
                ObjectReference(source=source, target=target),
            ]
        ) == {"object_reference:add": 1}

        refs = swh_storage.object_find_recent_references(target_swhid=target, limit=10)
        assert refs == [source]

        time.sleep(5)

        refs = swh_storage.object_find_recent_references(target_swhid=target, limit=10)
        assert refs == []



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import contextmanager
from functools import wraps
import logging
from time import monotonic
from types import TracebackType
from typing import Dict, Optional, Type

from swh.core.statsd import statsd

logger = logging.getLogger(__name__)

OPERATIONS_METRIC = "swh_storage_operations_total"
OPERATIONS_UNIT_METRIC = "swh_storage_operations_{unit}_total"
DURATION_METRIC = "swh_storage_request_duration_seconds"


def timed(f):
    """Time that function!"""

    @wraps(f)
    def d(*a, **kw):
        with statsd.timed(DURATION_METRIC, tags={"endpoint": f.__name__}):
            return f(*a, **kw)

    return d


class DifferentialTimer:
    """Compute differential timing metrics and send them to StatsD.

    The DifferentialTimer will send the ``metric`` to statsd, computing the
    difference between the time spent within the whole block, with the time
    spent within any blocks wrapped in the :meth:`inner` decorator.

    If an exception is raised,when sending the metric, the DifferentialTimer
    will add a ``inner_exc`` (when the exception is raised by an inner-timed
    section) or ``outer_exc`` tag with the name of the exception class that was
    raised .

    Arguments:
      metric: name of the timing metric that is sent to StatsD when the context
        manager exits
      tags: tags to attach to the metric when it is sent

    Example:
      To generate a ``metric_seconds`` timed metric, recording the overhead of
      running ``run_pre_processing`` and ``run_post_processing``, use:

      .. code-block:: python

          with DifferentialTimer("metric_seconds") as t:
              run_pre_processing()
              with t.inner():
                  run_inner_method()
              run_post_processing()

    """

    def __init__(self, metric: str, tags: Optional[Dict[str, str]] = None) -> None:
        self.start_ts: Optional[float] = None
        self.inner_elapsed = 0
        self.exc_source = "outer"
        self.metric = metric
        self.tags = tags or {}

    def __enter__(self) -> "DifferentialTimer":
        self.start_ts = monotonic()
        return self

    @contextmanager
    def inner(self):
        """Add the duration of this block to the inner elapsed time counter."""
        start = monotonic()
        try:
            yield
        except BaseException:
            self.exc_source = "inner"
            raise
        finally:
            self.inner_elapsed += monotonic() - start

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        _exc_value: Optional[BaseException],
        _traceback: Optional[TracebackType],
    ) -> None:
        if self.start_ts is None:
            logger.warning("DifferentialTimer exited without being started")
            return

        total_elapsed = monotonic() - self.start_ts

        extra_tags = {}
        if exc_type:
            extra_tags[f"{self.exc_source}_exc"] = exc_type.__name__

        if self.inner_elapsed == 0 and not exc_type:
            logger.warning("DifferentialTimer didn't record calls to `inner`?")

        # Statsd expects timings in milliseconds
        value = (total_elapsed - self.inner_elapsed) * 1000

        statsd.timing(
            self.metric,
            value,
            tags={**self.tags, **extra_tags},
        )


def send_metric(metric, count, method_name):
    """Send statsd metric with count for method `method_name`

    If count is 0, the metric is discarded.  If the metric is not
    parseable, the metric is discarded with a log message.

    Args:
        metric (str): Metric's name (e.g content:add, content:add:bytes)
        count (int): Associated value for the metric
        method_name (str): Method's name

    Returns:
        Bool to explicit if metric has been set or not
    """
    if count == 0:
        return False

    metric_type = metric.split(":")
    _length = len(metric_type)
    if _length == 2:
        object_type, operation = metric_type
        metric_name = OPERATIONS_METRIC
    elif _length == 3:
        object_type, operation, unit = metric_type
        metric_name = OPERATIONS_UNIT_METRIC.format(unit=unit)
    else:
        logging.warning("Skipping unknown metric {%s: %s}" % (metric, count))
        return False

    statsd.increment(
        metric_name,
        count,
        tags={
            "endpoint": method_name,
            "object_type": object_type,
            "operation": operation,
        },
    )
    return True


def process_metrics(f):
    """Increment object counters for the decorated function."""

    @wraps(f)
    def d(*a, **kw):
        r = f(*a, **kw)
        for metric, count in r.items():
            send_metric(metric=metric, count=count, method_name=f.__name__)

        return r

    return d



# Copyright (C) 2018-2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict, List

from attr import asdict, has
from confluent_kafka import Consumer
from hypothesis import given
from hypothesis.strategies import lists
import pytest

from swh.journal.pytest_plugin import assert_all_objects_consumed, consume_messages
from swh.model.hypothesis_strategies import objects
from swh.model.model import (
    Content,
    Directory,
    ExtID,
    MetadataAuthority,
    MetadataFetcher,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Person,
    RawExtrinsicMetadata,
    Release,
    Revision,
    SkippedContent,
    Snapshot,
)
from swh.model.tests.swh_model_data import TEST_OBJECTS
from swh.storage import get_storage

from ..in_memory import InMemoryStorage
from .storage_data import DEFAULT_ALGORITHMS, StorageData


def get_storage_with_writer(
    kafka_prefix: str, kafka_server: str, anonymize: bool
) -> InMemoryStorage:
    writer_config = {
        "cls": "kafka",
        "brokers": [kafka_server],
        "client_id": "kafka_writer",
        "prefix": kafka_prefix,
        "anonymize": anonymize,
        "auto_flush": False,
    }
    storage_config: Dict[str, Any] = {
        "cls": "pipeline",
        "steps": [
            {"cls": "memory", "journal_writer": writer_config},
        ],
    }

    ret = get_storage(**storage_config)
    assert isinstance(ret, InMemoryStorage)
    return ret


def test_storage_direct_writer(kafka_prefix: str, kafka_server, consumer: Consumer):

    storage = get_storage_with_writer(
        kafka_prefix=kafka_prefix, kafka_server=kafka_server, anonymize=False
    )

    object_types = (
        Content.object_type,
        SkippedContent.object_type,
        Directory.object_type,
        ExtID.object_type,
        MetadataAuthority.object_type,
        MetadataFetcher.object_type,
        Revision.object_type,
        Release.object_type,
        Snapshot.object_type,
        Origin.object_type,
        OriginVisit.object_type,
        OriginVisitStatus.object_type,
        RawExtrinsicMetadata.object_type,
    )

    expected_messages = 0

    for obj_type, objs in TEST_OBJECTS.items():
        method = getattr(storage, f"{obj_type}_add")
        if obj_type in object_types:
            method(objs)
            expected_messages += len(objs)
        else:
            assert False, obj_type
    storage.journal_writer.flush()

    existing_topics = set(
        topic
        for topic in consumer.list_topics(timeout=10).topics.keys()
        if topic.startswith(f"{kafka_prefix}.")  # final . to exclude privileged topics
    )
    assert existing_topics == {
        f"{kafka_prefix}.{obj_type}" for obj_type in object_types
    }

    consumed_messages = consume_messages(consumer, kafka_prefix, expected_messages)
    assert_all_objects_consumed(consumed_messages)


@pytest.mark.parametrize(
    "object_types",
    (set(k.value for k in TEST_OBJECTS.keys()) | {"hash_colliding_content"},),
)  # consume from the hash_colliding_content topic, which is not listened to by default
@pytest.mark.parametrize("colliding_hash", sorted(DEFAULT_ALGORITHMS))
def test_storage_direct_writer_content_hash_collision_sequential(
    kafka_prefix: str,
    kafka_server,
    consumer: Consumer,
    sample_data: StorageData,
    colliding_hash: str,
):
    storage = get_storage_with_writer(
        kafka_prefix=kafka_prefix, kafka_server=kafka_server, anonymize=False
    )
    for content in sample_data.colliding_contents[colliding_hash]:
        storage.content_add([content])
    assert storage.journal_writer.journal is not None
    storage.journal_writer.journal.flush()

    expected_messages = 4

    consumed_messages = consume_messages(consumer, kafka_prefix, expected_messages)
    colliding_no_ctime = [
        content.evolve(ctime=None)
        for content in sample_data.colliding_contents[colliding_hash]
    ]

    for msg in consumed_messages["hash_colliding_contents"]:
        assert Content.from_dict(msg[1]) in colliding_no_ctime


@pytest.mark.parametrize(
    "object_types",
    (set(k.value for k in TEST_OBJECTS.keys()) | {"hash_colliding_content"},),
)  # consume from the hash_colliding_content topic, which is not listened to by default
def test_storage_direct_writer_content_hash_collision_same_batch(
    kafka_prefix: str,
    kafka_server,
    consumer: Consumer,
    sample_data: StorageData,
):
    storage = get_storage_with_writer(
        kafka_prefix=kafka_prefix, kafka_server=kafka_server, anonymize=False
    )
    all_colliding: List[Content] = sum(sample_data.colliding_contents.values(), [])
    storage.content_add(all_colliding)
    assert storage.journal_writer.journal is not None
    storage.journal_writer.journal.flush()

    expected_messages = 2 * len(all_colliding)

    consumed_messages = consume_messages(consumer, kafka_prefix, expected_messages)
    colliding_no_ctime = [content.evolve(ctime=None) for content in all_colliding]

    for msg in consumed_messages["hash_colliding_contents"]:
        assert Content.from_dict(msg[1]) in colliding_no_ctime


def test_storage_direct_writer_anonymized(
    kafka_prefix: str, kafka_server, consumer: Consumer
):
    storage = get_storage_with_writer(
        kafka_prefix=kafka_prefix, kafka_server=kafka_server, anonymize=True
    )

    expected_messages = 0

    for obj_type, objs in TEST_OBJECTS.items():
        if obj_type == OriginVisit.object_type:
            # these have non-consistent API and are unrelated with what we
            # want to test here
            continue
        method = getattr(storage, f"{obj_type}_add")
        method(objs)
        expected_messages += len(objs)
    storage.journal_writer.flush()

    existing_topics = set(
        topic
        for topic in consumer.list_topics(timeout=10).topics.keys()
        if topic.startswith(kafka_prefix)
    )
    assert existing_topics == {
        f"{kafka_prefix}.{obj_type}"
        for obj_type in (
            "content",
            "directory",
            "extid",
            "metadata_authority",
            "metadata_fetcher",
            "origin",
            "origin_visit",
            "origin_visit_status",
            "raw_extrinsic_metadata",
            "release",
            "revision",
            "snapshot",
            "skipped_content",
        )
    } | {
        f"{kafka_prefix}_privileged.{obj_type}"
        for obj_type in (
            "release",
            "revision",
        )
    }


def check_anonymized_obj(obj):
    if has(obj):
        if isinstance(obj, Person):
            assert obj.name is None
            assert obj.email is None
            assert len(obj.fullname) == 32
        else:
            for key, value in asdict(obj, recurse=False).items():
                check_anonymized_obj(value)


@given(lists(objects(split_content=True)))
def test_anonymizer(obj_type_and_objs):
    for obj_type, obj in obj_type_and_objs:
        check_anonymized_obj(obj.anonymize())



# Copyright (C) 2019-2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import Counter
from typing import Optional
from unittest.mock import Mock

import attr
import pytest

from swh.storage import get_storage
from swh.storage.proxies.buffer import (
    BufferingProxyStorage,
    estimate_release_size,
    estimate_revision_size,
)
from swh.storage.utils import now


def get_storage_with_buffer_config(**buffer_config) -> BufferingProxyStorage:
    steps = [
        {"cls": "buffer", **buffer_config},
        {"cls": "memory"},
    ]

    ret = get_storage("pipeline", steps=steps)
    assert isinstance(ret, BufferingProxyStorage)
    return ret


def test_buffering_proxy_storage_content_threshold_not_hit(sample_data) -> None:
    contents = sample_data.contents[:2]
    contents_dict = [c.to_dict() for c in contents]

    storage = get_storage_with_buffer_config(
        min_batch_size={
            "content": 10,
        }
    )
    s = storage.content_add(contents)
    assert s == {}

    # contents have not been written to storage
    missing_contents = storage.content_missing(contents_dict)
    assert set(missing_contents) == set([contents[0].sha1, contents[1].sha1])

    s = storage.flush()
    assert s == {
        "content:add": 1 + 1,
        "content:add:bytes": contents[0].length + contents[1].length,
    }

    missing_contents = storage.content_missing(contents_dict)
    assert list(missing_contents) == []


def test_buffering_proxy_storage_content_threshold_nb_hit(sample_data) -> None:
    content = sample_data.content
    content_dict = content.to_dict()

    storage = get_storage_with_buffer_config(
        min_batch_size={
            "content": 1,
        }
    )

    s = storage.content_add([content])
    assert s == {
        "content:add": 1,
        "content:add:bytes": content.length,
    }

    missing_contents = storage.content_missing([content_dict])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_content_deduplicate(sample_data) -> None:
    contents = sample_data.contents[:2]
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "content": 2,
        }
    )

    s = storage.content_add([contents[0], contents[0]])
    assert s == {}

    s = storage.content_add([contents[0]])
    assert s == {}

    s = storage.content_add([contents[1]])
    assert s == {
        "content:add": 1 + 1,
        "content:add:bytes": contents[0].length + contents[1].length,
    }

    missing_contents = storage.content_missing([c.to_dict() for c in contents])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_content_metadata_deduplicate(sample_data) -> None:
    contents = [attr.evolve(cnt, ctime=now()) for cnt in sample_data.contents[:2]]

    storage = get_storage_with_buffer_config(
        min_batch_size={
            "content_metadata": 2,
        }
    )

    s = storage.content_add_metadata([contents[0], contents[0]])
    assert s == {}

    s = storage.content_add_metadata([contents[0]])
    assert s == {}

    s = storage.content_add_metadata([contents[1]])
    assert s == {
        "content:add": 1 + 1,
    }

    missing_contents = storage.content_missing([c.to_dict() for c in contents])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_content_threshold_bytes_hit(sample_data) -> None:
    contents = sample_data.contents[:2]
    content_bytes_min_batch_size = 2
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "content": 10,
            "content_bytes": content_bytes_min_batch_size,
        }
    )

    assert contents[0].length > content_bytes_min_batch_size

    s = storage.content_add([contents[0]])
    assert s == {
        "content:add": 1,
        "content:add:bytes": contents[0].length,
    }

    missing_contents = storage.content_missing([contents[0].to_dict()])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_skipped_content_threshold_not_hit(sample_data) -> None:
    contents = sample_data.skipped_contents
    contents_dict = [c.to_dict() for c in contents]
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "skipped_content": 10,
        }
    )
    s = storage.skipped_content_add([contents[0], contents[1]])
    assert s == {}

    # contents have not been written to storage
    missing_contents = storage.skipped_content_missing(contents_dict)
    assert {c["sha1"] for c in missing_contents} == {c.sha1 for c in contents}

    s = storage.flush()
    assert s == {"skipped_content:add": 1 + 1}

    missing_contents = storage.skipped_content_missing(contents_dict)
    assert list(missing_contents) == []


def test_buffering_proxy_storage_skipped_content_threshold_nb_hit(sample_data) -> None:
    contents = sample_data.skipped_contents
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "skipped_content": 1,
        }
    )

    s = storage.skipped_content_add([contents[0]])
    assert s == {"skipped_content:add": 1}

    missing_contents = storage.skipped_content_missing([contents[0].to_dict()])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_skipped_content_deduplicate(sample_data):
    contents = sample_data.skipped_contents[:2]
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "skipped_content": 2,
        }
    )

    s = storage.skipped_content_add([contents[0], contents[0]])
    assert s == {}

    s = storage.skipped_content_add([contents[0]])
    assert s == {}

    s = storage.skipped_content_add([contents[1]])
    assert s == {
        "skipped_content:add": 1 + 1,
    }

    missing_contents = storage.skipped_content_missing([c.to_dict() for c in contents])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_extid_threshold_not_hit(sample_data) -> None:
    extid = sample_data.extid1
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "extid": 10,
        }
    )
    s = storage.extid_add([extid])
    assert s == {}

    present_extids = storage.extid_get_from_target(
        extid.target.object_type, [extid.target.object_id]
    )
    assert list(present_extids) == []

    s = storage.flush()
    assert s == {
        "extid:add": 1,
    }

    present_extids = storage.extid_get_from_target(
        extid.target.object_type, [extid.target.object_id]
    )
    assert list(present_extids) == [extid]


def test_buffering_proxy_storage_extid_threshold_hit(sample_data) -> None:
    extid = sample_data.extid1
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "extid": 1,
        }
    )
    s = storage.extid_add([extid])
    assert s == {
        "extid:add": 1,
    }

    present_extids = storage.extid_get_from_target(
        extid.target.object_type, [extid.target.object_id]
    )
    assert list(present_extids) == [extid]

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_extid_deduplicate(sample_data) -> None:
    extids = sample_data.extids[:2]
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "extid": 2,
        }
    )

    s = storage.extid_add([extids[0], extids[0]])
    assert s == {}

    s = storage.extid_add([extids[0]])
    assert s == {}

    s = storage.extid_add([extids[1]])
    assert s == {
        "extid:add": 1 + 1,
    }

    for extid in extids:
        present_extids = storage.extid_get_from_target(
            extid.target.object_type, [extid.target.object_id]
        )
        assert list(present_extids) == [extid]

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_directory_threshold_not_hit(sample_data) -> None:
    directory = sample_data.directory
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "directory": 10,
        }
    )
    s = storage.directory_add([directory])
    assert s == {}

    missing_directories = storage.directory_missing([directory.id])
    assert list(missing_directories) == [directory.id]

    s = storage.flush()
    assert s == {
        "directory:add": 1,
    }

    missing_directories = storage.directory_missing([directory.id])
    assert list(missing_directories) == []


def test_buffering_proxy_storage_directory_threshold_hit(sample_data) -> None:
    directory = sample_data.directory
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "directory": 1,
        }
    )
    s = storage.directory_add([directory])
    assert s == {
        "directory:add": 1,
    }

    missing_directories = storage.directory_missing([directory.id])
    assert list(missing_directories) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_directory_deduplicate(sample_data) -> None:
    directories = sample_data.directories[:2]
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "directory": 2,
        }
    )

    s = storage.directory_add([directories[0], directories[0]])
    assert s == {}

    s = storage.directory_add([directories[0]])
    assert s == {}

    s = storage.directory_add([directories[1]])
    assert s == {
        "directory:add": 1 + 1,
    }

    missing_directories = storage.directory_missing([d.id for d in directories])
    assert list(missing_directories) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_directory_entries_threshold(sample_data) -> None:
    directories = sample_data.directories
    n_entries = sum(len(d.entries) for d in directories)
    threshold = sum(len(d.entries) for d in directories[:-2])

    # ensure the threshold is in the middle
    assert 0 < threshold < n_entries

    storage = get_storage_with_buffer_config(
        min_batch_size={"directory_entries": threshold}
    )
    storage.storage = Mock(wraps=storage.storage)

    for directory in directories:
        storage.directory_add([directory])
    storage.flush()

    # We should have called the underlying directory_add at least twice, as
    # we have hit the threshold for number of entries on directory n-2
    method_calls = Counter(c[0] for c in storage.storage.method_calls)
    assert method_calls["directory_add"] >= 2


def test_buffering_proxy_storage_revision_threshold_not_hit(sample_data) -> None:
    revision = sample_data.revision
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "revision": 10,
        }
    )
    s = storage.revision_add([revision])
    assert s == {}

    missing_revisions = storage.revision_missing([revision.id])
    assert list(missing_revisions) == [revision.id]

    s = storage.flush()
    assert s == {
        "revision:add": 1,
    }

    missing_revisions = storage.revision_missing([revision.id])
    assert list(missing_revisions) == []


def test_buffering_proxy_storage_revision_threshold_hit(sample_data) -> None:
    revision = sample_data.revision
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "revision": 1,
        }
    )
    s = storage.revision_add([revision])
    assert s == {
        "revision:add": 1,
    }

    missing_revisions = storage.revision_missing([revision.id])
    assert list(missing_revisions) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_revision_deduplicate(sample_data) -> None:
    revisions = sample_data.revisions[:2]
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "revision": 2,
        }
    )

    s = storage.revision_add([revisions[0], revisions[0]])
    assert s == {}

    s = storage.revision_add([revisions[0]])
    assert s == {}

    s = storage.revision_add([revisions[1]])
    assert s == {
        "revision:add": 1 + 1,
    }

    missing_revisions = storage.revision_missing([r.id for r in revisions])
    assert list(missing_revisions) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_revision_parents_threshold(sample_data) -> None:
    revisions = sample_data.revisions
    n_parents = sum(len(r.parents) for r in revisions)
    threshold = sum(len(r.parents) for r in revisions[:-2])

    # ensure the threshold is in the middle
    assert 0 < threshold < n_parents

    storage = get_storage_with_buffer_config(
        min_batch_size={"revision_parents": threshold}
    )
    storage.storage = Mock(wraps=storage.storage)

    for revision in revisions:
        storage.revision_add([revision])
    storage.flush()

    # We should have called the underlying revision_add at least twice, as
    # we have hit the threshold for number of parents on revision n-2
    method_calls = Counter(c[0] for c in storage.storage.method_calls)
    assert method_calls["revision_add"] >= 2


def test_buffering_proxy_storage_revision_size_threshold(sample_data) -> None:
    revisions = sample_data.revisions
    total_size = sum(estimate_revision_size(r) for r in revisions)
    threshold = sum(estimate_revision_size(r) for r in revisions[:-2])

    # ensure the threshold is in the middle
    assert 0 < threshold < total_size

    storage = get_storage_with_buffer_config(
        min_batch_size={"revision_bytes": threshold}
    )
    storage.storage = Mock(wraps=storage.storage)

    for revision in revisions:
        storage.revision_add([revision])
    storage.flush()

    # We should have called the underlying revision_add at least twice, as
    # we have hit the threshold for number of parents on revision n-2
    method_calls = Counter(c[0] for c in storage.storage.method_calls)
    assert method_calls["revision_add"] >= 2


def test_buffering_proxy_storage_release_threshold_not_hit(sample_data) -> None:
    releases = sample_data.releases
    threshold = 10

    assert len(releases) < threshold
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "release": threshold,
        }  # configuration set
    )
    s = storage.release_add(releases)
    assert s == {}

    release_ids = [r.id for r in releases]
    missing_releases = storage.release_missing(release_ids)
    assert list(missing_releases) == release_ids

    s = storage.flush()
    assert s == {
        "release:add": len(releases),
    }

    missing_releases = storage.release_missing(release_ids)
    assert list(missing_releases) == []


def test_buffering_proxy_storage_release_threshold_hit(sample_data) -> None:
    releases = sample_data.releases
    threshold = 2
    assert len(releases) > threshold

    storage = get_storage_with_buffer_config(
        min_batch_size={
            "release": threshold,
        }  # configuration set
    )

    s = storage.release_add(releases)
    assert s == {
        "release:add": len(releases),
    }

    release_ids = [r.id for r in releases]
    missing_releases = storage.release_missing(release_ids)
    assert list(missing_releases) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_release_deduplicate(sample_data) -> None:
    releases = sample_data.releases[:2]
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "release": 2,
        }
    )

    s = storage.release_add([releases[0], releases[0]])
    assert s == {}

    s = storage.release_add([releases[0]])
    assert s == {}

    s = storage.release_add([releases[1]])
    assert s == {
        "release:add": 1 + 1,
    }

    missing_releases = storage.release_missing([r.id for r in releases])
    assert list(missing_releases) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_release_size_threshold(sample_data) -> None:
    releases = sample_data.releases
    total_size = sum(estimate_release_size(r) for r in releases)
    threshold = sum(estimate_release_size(r) for r in releases[:-2])

    # ensure the threshold is in the middle
    assert 0 < threshold < total_size

    storage = get_storage_with_buffer_config(
        min_batch_size={"release_bytes": threshold}
    )
    storage.storage = Mock(wraps=storage.storage)

    for release in releases:
        storage.release_add([release])
    storage.flush()

    # We should have called the underlying release_add at least twice, as
    # we have hit the threshold for number of parents on release n-2
    method_calls = Counter(c[0] for c in storage.storage.method_calls)
    assert method_calls["release_add"] >= 2


def test_buffering_proxy_storage_snapshot_threshold_not_hit(sample_data) -> None:
    snapshots = sample_data.snapshots
    threshold = 10

    assert len(snapshots) < threshold
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "snapshot": threshold,
        }  # configuration set
    )
    s = storage.snapshot_add(snapshots)
    assert s == {}

    snapshot_ids = [r.id for r in snapshots]
    missing_snapshots = storage.snapshot_missing(snapshot_ids)
    assert list(missing_snapshots) == snapshot_ids

    s = storage.flush()
    assert s == {
        "snapshot:add": len(snapshots),
    }

    missing_snapshots = storage.snapshot_missing(snapshot_ids)
    assert list(missing_snapshots) == []


def test_buffering_proxy_storage_snapshot_threshold_hit(sample_data) -> None:
    snapshots = sample_data.snapshots
    threshold = 2
    assert len(snapshots) > threshold

    storage = get_storage_with_buffer_config(
        min_batch_size={
            "snapshot": threshold,
        }  # configuration set
    )

    s = storage.snapshot_add(snapshots)
    assert s == {
        "snapshot:add": len(snapshots),
    }

    snapshot_ids = [r.id for r in snapshots]
    missing_snapshots = storage.snapshot_missing(snapshot_ids)
    assert list(missing_snapshots) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_snapshot_deduplicate(sample_data) -> None:
    snapshots = sample_data.snapshots[:2]
    storage = get_storage_with_buffer_config(
        min_batch_size={
            "snapshot": 2,
        }
    )

    s = storage.snapshot_add([snapshots[0], snapshots[0]])
    assert s == {}

    s = storage.snapshot_add([snapshots[0]])
    assert s == {}

    s = storage.snapshot_add([snapshots[1]])
    assert s == {
        "snapshot:add": 1 + 1,
    }

    missing_snapshots = storage.snapshot_missing([r.id for r in snapshots])
    assert list(missing_snapshots) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_clear(sample_data) -> None:
    """Clear operation on buffer"""
    threshold = 10
    contents = sample_data.contents
    assert 0 < len(contents) < threshold
    skipped_contents = sample_data.skipped_contents
    assert 0 < len(skipped_contents) < threshold
    directories = sample_data.directories
    assert 0 < len(directories) < threshold
    revisions = sample_data.revisions
    assert 0 < len(revisions) < threshold
    releases = sample_data.releases
    assert 0 < len(releases) < threshold
    snapshots = sample_data.snapshots
    assert 0 < len(snapshots) < threshold
    metadata = sample_data.content_metadata + sample_data.origin_metadata
    assert 0 < len(metadata) < threshold

    storage = get_storage_with_buffer_config(
        min_batch_size={
            "content": threshold,
            "skipped_content": threshold,
            "directory": threshold,
            "revision": threshold,
            "release": threshold,
            "raw_extrinsic_metadata": threshold,
        }
    )

    storage.metadata_authority_add(sample_data.authorities)
    storage.metadata_fetcher_add(sample_data.fetchers)

    s = storage.content_add(contents)
    assert s == {}
    s = storage.skipped_content_add(skipped_contents)
    assert s == {}
    s = storage.directory_add(directories)
    assert s == {}
    s = storage.revision_add(revisions)
    assert s == {}
    s = storage.release_add(releases)
    assert s == {}
    s = storage.snapshot_add(snapshots)
    assert s == {}
    s = storage.raw_extrinsic_metadata_add(metadata)
    assert s == {}

    assert len(storage._objects["content"]) == len(contents)
    assert len(storage._objects["skipped_content"]) == len(skipped_contents)
    assert len(storage._objects["directory"]) == len(directories)
    assert len(storage._objects["revision"]) == len(revisions)
    assert len(storage._objects["release"]) == len(releases)
    assert len(storage._objects["snapshot"]) == len(snapshots)
    assert len(storage._objects["raw_extrinsic_metadata"]) == len(metadata)

    # clear only content from the buffer
    s = storage.clear_buffers(["content"])  # type: ignore
    assert s is None

    # specific clear operation on specific object type content only touched
    # them
    assert len(storage._objects["content"]) == 0
    assert len(storage._objects["skipped_content"]) == len(skipped_contents)
    assert len(storage._objects["directory"]) == len(directories)
    assert len(storage._objects["revision"]) == len(revisions)
    assert len(storage._objects["release"]) == len(releases)
    assert len(storage._objects["snapshot"]) == len(snapshots)
    assert len(storage._objects["raw_extrinsic_metadata"]) == len(metadata)

    # clear current buffer from all object types
    storage.clear_buffers()

    assert len(storage._objects["content"]) == 0
    assert len(storage._objects["skipped_content"]) == 0
    assert len(storage._objects["directory"]) == 0
    assert len(storage._objects["revision"]) == 0
    assert len(storage._objects["release"]) == 0
    assert len(storage._objects["snapshot"]) == 0
    assert len(storage._objects["raw_extrinsic_metadata"]) == 0


def test_buffer_proxy_with_default_args() -> None:
    storage = get_storage_with_buffer_config()

    assert storage is not None


def test_buffer_flush_stats(sample_data) -> None:
    storage = get_storage_with_buffer_config()

    storage.metadata_authority_add(sample_data.authorities)
    storage.metadata_fetcher_add(sample_data.fetchers)

    s = storage.content_add(sample_data.contents)
    assert s == {}
    s = storage.skipped_content_add(sample_data.skipped_contents)
    assert s == {}
    s = storage.directory_add(sample_data.directories)
    assert s == {}
    s = storage.revision_add(sample_data.revisions)
    assert s == {}
    s = storage.release_add(sample_data.releases)
    assert s == {}
    s = storage.snapshot_add(sample_data.snapshots)
    assert s == {}
    s = storage.raw_extrinsic_metadata_add(
        sample_data.content_metadata + sample_data.origin_metadata
    )
    assert s == {}

    # Flush all the things
    s = storage.flush()
    assert s["content:add"] > 0
    assert s["content:add:bytes"] > 0
    assert s["skipped_content:add"] > 0
    assert s["directory:add"] > 0
    assert s["revision:add"] > 0
    assert s["release:add"] > 0
    assert s["snapshot:add"] > 0
    assert s["cnt_metadata:add"] > 0
    assert s["ori_metadata:add"] > 0


def test_buffer_operation_order(sample_data) -> None:
    storage = get_storage_with_buffer_config()

    # not buffered (because they happen rarely)
    storage.metadata_authority_add(sample_data.authorities)
    storage.metadata_fetcher_add(sample_data.fetchers)

    # Wrap the inner storage in a mock to track all method calls.
    storage.storage = mocked_storage = Mock(wraps=storage.storage)

    # Simulate a loader: add origin metadata, contents, directories, revisions,
    # releases, snapshots, then content metadata.
    storage.raw_extrinsic_metadata_add(sample_data.origin_metadata)
    storage.content_add(sample_data.contents)
    storage.skipped_content_add(sample_data.skipped_contents)
    storage.directory_add(sample_data.directories)
    storage.revision_add(sample_data.revisions)
    storage.release_add(sample_data.releases)
    storage.snapshot_add(sample_data.snapshots)
    storage.raw_extrinsic_metadata_add(sample_data.content_metadata)

    # Check that nothing has been flushed yet
    assert mocked_storage.method_calls == []

    # Flush all the things
    storage.flush()

    methods_called = [c[0] for c in mocked_storage.method_calls]
    prev = -1
    for method in [
        "raw_extrinsic_metadata_add",
        "content_add",
        "skipped_content_add",
        "directory_add",
        "revision_add",
        "release_add",
        "snapshot_add",
        "flush",
    ]:
        try:
            cur: Optional[int] = methods_called.index(method)
        except ValueError:
            cur = None

        assert cur is not None, "Method %s not called" % method
        assert cur > prev, "Method %s called out of order; all calls were: %s" % (
            method,
            methods_called,
        )
        prev = cur


def test_buffer_empty_batches() -> None:
    "Flushing an empty buffer storage doesn't call any underlying _add method"
    storage = get_storage_with_buffer_config()
    storage.storage = mocked_storage = Mock(wraps=storage.storage)

    storage.flush()
    methods_called = {c[0] for c in mocked_storage.method_calls}
    assert methods_called == {"flush", "clear_buffers"}


def test_buffer_warning_when_flush_is_missing_flush_ok(sample_data) -> None:
    storage = get_storage_with_buffer_config()
    storage.content_add(sample_data.contents)
    # storage.flush() is intentionally missing

    with pytest.warns(UserWarning, match="during shutdown. A call"):
        del storage


def test_buffer_warning_when_flush_is_missing_flush_fails(mocker, sample_data) -> None:
    storage = get_storage_with_buffer_config()
    storage.content_add(sample_data.contents)
    # storage.flush() is intentionally missing

    # Make underlying flush() on the underlying storage crash
    # to test if we get the right warning.
    mocker.patch.object(storage.storage, "flush", side_effect=RuntimeError("KO"))

    with pytest.warns(UserWarning, match="They are now probably lost"):
        del storage



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import csv
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    TextIO,
    Tuple,
)
import warnings

import click

from ...cli import storage as storage_cli_group

if TYPE_CHECKING:
    from swh.model.swhids import ExtendedSWHID

    from .db import MaskedState, MaskingRequest


def storage_config_iter(storage_config: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
    yield storage_config
    if storage_config["cls"] == "pipeline":
        for step in storage_config["steps"]:
            yield step
    elif "storage" in storage_config:
        yield storage_config["storage"]


@storage_cli_group.group(name="masking")
@click.pass_context
def masking_cli_group(ctx: click.Context) -> click.Context:
    """Configure masking on archived objects

    These tools require read/write access to the masking database.
    An entry must be added to the configuration file as follow::

        \b
        storage:
          …
        \b
        masking_admin:
          cls: postgresql
          db: "service=swh-masking-admin"
    """

    if "masking_admin" not in ctx.obj["config"] or (
        "masking_db" not in ctx.obj["config"]["masking_admin"]
        and "db" not in ctx.obj["config"]["masking_admin"]
    ):
        ctx.fail(
            "You must have a masking_admin, with a db entry, "
            "configured in your config file."
        )

    if "masking_db" in ctx.obj["config"]["masking_admin"]:
        warnings.warn(
            "Please use the db field for the masking admin configuration, which "
            "makes it compatible with the `swh db` command-line utilities",
            DeprecationWarning,
        )

    from psycopg import OperationalError

    from .db import MaskingAdmin

    try:
        db = None
        for key in ("db", "masking_db"):
            db = ctx.obj["config"]["masking_admin"].get(key)
            if db:
                break
        if not db:
            assert False, "Existence of config entries has been checked earlier"

        ctx.obj["masking_admin"] = MaskingAdmin.connect(db)
    except OperationalError as ex:
        raise click.ClickException(str(ex))

    return ctx


class EditAborted(BaseException):
    pass


def edit_message(prompt: str, extra_lines: List[str] = []):
    """Edit a message through click.edit() adding some extra context, filtering
    comments, and raising an exception on empty messages"""

    message_base = (
        "\n\n"
        f"# {prompt}\n"
        "# Lines starting with “#” will be ignored. "
        "An empty message will abort the operation.\n"
    )
    if extra_lines:
        message_base += "#\n# " + "\n# ".join(extra_lines)
    message = click.edit(message_base)
    if message is None:
        raise EditAborted()
    message = "\n".join(
        line.strip()
        for line in message.lstrip().split("\n")
        if not line.startswith("#")
    ).rstrip()
    if message == "":
        raise EditAborted()
    return message


class RequestType(click.ParamType):
    name = "request slug or uuid"

    def convert(self, value, param, ctx) -> "MaskingRequest":
        # Try from UUID first
        from uuid import UUID

        try:
            uuid = UUID(value)
        except ValueError:
            pass
        else:
            request = ctx.obj["masking_admin"].find_request_by_id(uuid)
            if not request:
                raise click.ClickException(f"Request “{uuid}” not found from id.")
            return request

        # Try from slug
        request = ctx.obj["masking_admin"].find_request(value)
        if not request:
            raise click.ClickException(f"Request “{value}” not found.")
        return request


class SWHIDType(click.ParamType):
    name = "swhid"

    def convert(self, value, param, ctx) -> "ExtendedSWHID":
        from swh.model.swhids import ExtendedSWHID, ValidationError

        try:
            return ExtendedSWHID.from_string(value)
        except ValidationError:
            raise click.ClickException(f"Unable to parse “{value}” as a SWHID.")


@masking_cli_group.command(name="new-request")
@click.option(
    "-m", "--message", "reason", metavar="REASON", help="why the request was made"
)
@click.argument("slug")
@click.pass_context
def new_request(ctx: click.Context, slug: str, reason: Optional[str] = None) -> None:
    """Create a new request to mask objects

    SLUG is a human-readable unique identifier for the request. It is an
    internal identifier that will be used in subsequent commands to address this
    newly recorded request.

    A reason for the request must be specified, either using the `-m` option or
    via the provided editor.
    """

    from .db import DuplicateRequest

    if reason is None or reason == "":
        try:
            reason = edit_message(f"Please enter a reason for the request “{slug}”.")
        except EditAborted:
            raise click.ClickException("Aborting due to an empty reason.")

    try:
        request = ctx.obj["masking_admin"].create_request(slug, reason)
    except DuplicateRequest:
        raise click.ClickException(f"Request “{slug}” already exists.")
    click.echo(
        f"New request “{click.style(slug, fg='green', bold=True)}” "
        f"recorded with identifier {request.id}"
    )


@masking_cli_group.command(name="list-requests")
@click.option(
    "-a",
    "--include-cleared-requests/--exclude-cleared-requests",
    default=False,
    help="Show requests without any masking state",
)
@click.pass_context
def list_requests(ctx: click.Context, include_cleared_requests: bool) -> None:
    """List masking requests"""

    def list_output() -> Iterator[str]:
        for request, mask_count in ctx.obj["masking_admin"].get_requests(
            include_cleared_requests=include_cleared_requests
        ):
            yield f"📄 {click.style(request.slug, bold=True)}\n"
            yield f"📊 {mask_count} object{'s' if mask_count != 1 else ''}\n"
            yield f"🪪 {request.id}\n"
            # XXX: humanize would be nice here as well
            yield f"📅 {request.date.strftime('%c %z')}\n"
            yield from output_message(request.reason)
            yield "\n"

    click.echo_via_pager(list_output())


class MaskedStateType(click.Choice):
    def __init__(self):
        super().__init__(
            ["visible", "decision-pending", "restricted"], case_sensitive=False
        )


def parse_masked_state(str: str) -> "MaskedState":
    from .db import MaskedState

    return MaskedState[str.replace("-", "_").upper()]


def format_masked_state(state: "MaskedState") -> str:
    return state.name.lower().replace("_", "-")


def read_swhids(file: TextIO) -> List["ExtendedSWHID"]:
    import re

    from swh.model.swhids import ExtendedSWHID

    filter_re = re.compile(r"^(#|$)")
    return [
        ExtendedSWHID.from_string(line.strip())
        for line in file.read().split("\n")
        if not filter_re.match(line)
    ]


@masking_cli_group.command(name="update-objects")
@click.option("-m", "--message", help="an explanation for this change")
@click.option("-f", "--file", type=click.File(), help="a file with on SWHID per line")
@click.argument("request", metavar="SLUG", type=RequestType())
@click.argument("new-state", metavar="NEW_STATE", type=MaskedStateType())
@click.pass_context
def update_objects(
    ctx: click.Context,
    request: "MaskingRequest",
    new_state: str,
    message: Optional[str] = None,
    file: Optional[TextIO] = None,
) -> None:
    """Update the state of given objects

    The masked state of the provided SWHIDs will be updated to NEW_STATE for the
    request SLUG.

    NEW_STATE must be one of “visible”, “decision-pending” or “restricted”.

    SWHIDs must be provided one per line, either via the standard input or a
    file specified via the `-f` option. `-` is synonymous for the standard
    input.

    An explanation for this change must be added to the request history. It can
    either be specified by the `-m` option or via the provided editor.
    """

    import sys

    from swh.model.swhids import ValidationError

    from .db import RequestNotFound

    if file is None:
        file = sys.stdin

    masked_state = parse_masked_state(new_state)
    try:
        swhids = read_swhids(file)
    except ValidationError as ex:
        raise click.ClickException(ex.messages[0])

    if len(swhids) == 0:
        raise click.ClickException("No SWHIDs given!")

    if message is None or message == "":
        try:
            message = edit_message(
                f"Please enter an explanation for this update on request “{request.slug}”."
            )
        except EditAborted:
            raise click.ClickException("Aborting due to an empty explanation.")

    try:
        ctx.obj["masking_admin"].set_object_state(request.id, masked_state, swhids)
        ctx.obj["masking_admin"].record_history(request.id, message)
    except RequestNotFound:
        raise click.ClickException(f"Request with id “{request.id}” not found.")

    click.echo(
        f"Updated masking state for {len(swhids)} objects in request “{request.slug}”."
    )


@masking_cli_group.command(name="status")
@click.argument("request", metavar="SLUG", type=RequestType())
@click.pass_context
def status_cmd(ctx: click.Context, request: "MaskingRequest") -> None:
    """Get the masking states defined by a request"""

    states = ctx.obj["masking_admin"].get_states_for_request(request.id)
    for swhid, state in states.items():
        click.echo(f"{swhid} {format_masked_state(state)}")


def output_message(message: str) -> Iterator[str]:
    import textwrap

    for i, line in enumerate(message.splitlines()):
        yield textwrap.fill(
            line,
            initial_indent="🗩  " if i == 0 else "   ",
            subsequent_indent="    ",
            replace_whitespace=False,
        )
        yield "\n"


@masking_cli_group.command(name="history")
@click.argument("request", metavar="SLUG", type=RequestType())
@click.pass_context
def history_cmd(ctx: click.Context, request: "MaskingRequest") -> None:
    """Get the history for a request"""

    def history_output() -> Iterator[str]:
        # XXX: we have no way to tie a record to a set of swhids or their state
        # the only thing we can do is display the messages
        yield f"History for request “{request.slug}” ({request.id}) \n"
        yield from output_message(request.reason)
        yield "\n"
        for record in ctx.obj["masking_admin"].get_history(request.id):
            # XXX: if we agree to bring humanize in the dependencies
            # relative_time = humanize.naturaltime(
            #   dt.datetime.now(tz=dt.timezone.utc) - record.date
            # )
            yield f"📅 {record.date.strftime('%c %z')}\n"
            for i, line in enumerate(record.message.splitlines()):
                yield from output_message(record.message)

    click.echo_via_pager(history_output())


@masking_cli_group.command(name="object-state")
@click.argument("swhids", metavar="SWHID", nargs=-1, type=SWHIDType())
@click.pass_context
def object_state(ctx: click.Context, swhids: List["ExtendedSWHID"]) -> None:
    """Get the masking state for a set of SWHIDs

    If an object given in the arguments is not listed in the output,
    it means no masking state is set in any requests.
    """

    import itertools

    from .db import MaskedState

    STATE_TO_COLORS = {
        MaskedState.VISIBLE: "bright_green",
        MaskedState.DECISION_PENDING: "bright_yellow",
        MaskedState.RESTRICTED: "bright_red",
    }
    STATE_TO_LABEL = {
        MaskedState.VISIBLE: "visible",
        MaskedState.DECISION_PENDING: "decision-pending",
        MaskedState.RESTRICTED: "restricted",
    }

    def state_output() -> Iterator[str]:
        # find_masks() will group masks for the same SWHID
        for swhid, masks_iter in itertools.groupby(
            ctx.obj["masking_admin"].find_masks(swhids), key=lambda mask: mask.swhid
        ):
            masks = list(masks_iter)
            masked = any(mask.state != MaskedState.VISIBLE for mask in masks)
            yield (
                f"{'masked ' if masked else 'visible'} "
                f"{click.style(str(masks[0].swhid), bold=masked)}\n"
            )
            for mask in masks:
                yield click.style(
                    f"        {mask.request_slug}: {STATE_TO_LABEL[mask.state]}\n",
                    fg=STATE_TO_COLORS[mask.state],
                )

    click.echo_via_pager(state_output())


@masking_cli_group.command(name="clear-request")
@click.option("-m", "--message", help="an explanation for this change")
@click.argument("request", metavar="SLUG", type=RequestType())
@click.pass_context
def clear_request(
    ctx: click.Context, request: "MaskingRequest", message: Optional[str]
) -> None:
    """Remove all masking states for the given request"""

    from .db import RequestNotFound

    if message is None or message == "":
        states = ctx.obj["masking_admin"].get_states_for_request(request.id)
        extra_lines = ["Associated object states:", ""]
        for swhid, state in states.items():
            extra_lines.append(f"{swhid} {format_masked_state(state)}")
        try:
            message = edit_message(
                "Please enter an explanation for clearing masks"
                f" from request “{request.slug}”.",
                extra_lines=extra_lines,
            )
        except EditAborted:
            raise click.ClickException("Aborting due to an empty explanation.")

    try:
        ctx.obj["masking_admin"].delete_masks(request.id)
        ctx.obj["masking_admin"].record_history(request.id, message)
    except RequestNotFound:
        raise click.ClickException(f"Request with id “{request.id}” not found.")

    click.echo(f"Masks cleared for request “{request.slug}”.")


@masking_cli_group.group(name="patching")
@click.pass_context
def patching_cli_group(ctx: click.Context) -> click.Context:
    """Tools to manage the patching of objects"""
    return ctx


def read_display_names(file: Iterable) -> List[Tuple[bytes, bytes]]:
    data = list(csv.reader(file))
    assert all(len(x) == 2 for x in data)
    return [(name.encode(), newname.encode()) for (name, newname) in data]


@patching_cli_group.command(name="set")
@click.argument("input", type=click.File("r"))
@click.option(
    "--clear/--keep",
    help="Clear the display names table before inserting new entries",
)
@click.pass_context
def set_patching_entries(
    ctx: click.Context,
    input,
    clear: bool,
) -> None:
    """Set display names (patching entries)"""
    display_names = read_display_names(input)

    db = ctx.obj["masking_admin"]
    db.set_display_names(display_names, clear=clear)
    click.echo("Display names updated")



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import contextmanager
from typing import Dict, Iterable, Iterator, List, Optional, Union
import warnings

import psycopg_pool

from swh.model.model import Origin, OriginVisit, OriginVisitStatus
from swh.storage import get_storage
from swh.storage.exc import BlockedOriginException
from swh.storage.interface import StorageInterface
from swh.storage.metrics import DifferentialTimer
from swh.storage.proxies.blocking.db import BlockingState

from .db import BlockingQuery

BLOCKING_OVERHEAD_METRIC = "swh_storage_blocking_overhead_seconds"


def get_datastore(cls, db=None, blocking_db=None, **kwargs):
    assert cls in ("postgresql", "blocking")
    from .db import BlockingAdmin

    if db is None:
        db = blocking_db
    return BlockingAdmin.connect(db)


def blocking_overhead_timer(method_name: str) -> DifferentialTimer:
    """Return a properly setup DifferentialTimer for ``method_name`` of the storage"""
    return DifferentialTimer(BLOCKING_OVERHEAD_METRIC, tags={"endpoint": method_name})


class BlockingProxyStorage:
    """Blocking storage proxy

    This proxy prevents visits from a known list of origins to be performed at all.

    It uses a specific PostgreSQL database (which for now is colocated with the
    swh.storage PostgreSQL database), the access to which is implemented in the
    :mod:`..blocking.db` submodule.

    Sample configuration

    .. code-block: yaml

        storage:
          cls: blocking
          db: 'dbname=swh-blocking-proxy'
          max_pool_conns: 10
          storage:
          - cls: remote
            url: http://storage.internal.staging.swh.network:5002/

    """

    def __init__(
        self,
        storage: Union[Dict, StorageInterface],
        db: Optional[str] = None,
        blocking_db: Optional[str] = None,
        min_pool_conns: int = 1,
        max_pool_conns: int = 5,
    ):
        if db is None:
            assert blocking_db is not None
            warnings.warn(
                "'blocking_db' field in the blocking storage configuration "
                "was renamed 'db' field",
                DeprecationWarning,
            )
            db = blocking_db
        self.storage: StorageInterface = (
            get_storage(**storage) if isinstance(storage, dict) else storage
        )

        self._blocking_pool = psycopg_pool.ConnectionPool(
            db,
            min_size=min_pool_conns,
            max_size=max_pool_conns,
            open=True,
        )

    def origin_visit_status_add(
        self, visit_statuses: List[OriginVisitStatus]
    ) -> Dict[str, int]:
        with self._blocking_query() as q:
            statuses = q.origins_are_blocked([v.origin for v in visit_statuses])
            if statuses and any(
                status.state != BlockingState.NON_BLOCKED
                for status in statuses.values()
            ):
                raise BlockedOriginException(statuses)
        return self.storage.origin_visit_status_add(visit_statuses)

    def origin_visit_add(self, visits: List[OriginVisit]) -> Iterable[OriginVisit]:
        with self._blocking_query() as q:
            statuses = q.origins_are_blocked([v.origin for v in visits])
            if statuses and any(
                status.state != BlockingState.NON_BLOCKED
                for status in statuses.values()
            ):
                raise BlockedOriginException(statuses)
        return self.storage.origin_visit_add(visits)

    def origin_add(self, origins: List[Origin]) -> Dict[str, int]:
        with self._blocking_query() as q:
            statuses = {}
            for origin in origins:
                status = q.origin_is_blocked(origin.url)
                if status and status.state != BlockingState.NON_BLOCKED:
                    statuses[origin.url] = status
            if statuses:
                raise BlockedOriginException(statuses)
        return self.storage.origin_add(origins)

    @contextmanager
    def _blocking_query(self) -> Iterator[BlockingQuery]:
        ret = None
        try:
            ret = BlockingQuery.from_pool(self._blocking_pool)
            yield ret
        finally:
            if ret:
                ret.put_conn()

    def __getattr__(self, key):
        method = getattr(self.storage, key)
        if method:
            # Avoid going through __getattr__ again next time
            setattr(self, key, method)
            return method

        # Raise a NotImplementedError to make sure we don't forget to add
        # masking to any new storage functions
        raise NotImplementedError(key)



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import call
import uuid

import pytest

from swh.core.db.db_utils import get_database_info
from swh.storage.proxies.blocking.db import (
    BlockingAdmin,
    BlockingQuery,
    BlockingState,
    BlockingStatus,
    DuplicateRequest,
    RequestNotFound,
    get_urls_to_check,
)
from swh.storage.tests.storage_data import StorageData


@pytest.mark.parametrize(
    "url,exact,prefix",
    [
        pytest.param(
            "https://github.com/user1/test1.git",
            [
                "https://github.com/user1/test1.git",
                "https://github.com/user1/test1",
            ],
            ["https://github.com/user1", "https://github.com"],
            id="known-suffix-git",
        ),
        pytest.param(
            "https://foss.heptapod.net/user1/test1.hg",
            [
                "https://foss.heptapod.net/user1/test1.hg",
                "https://foss.heptapod.net/user1/test1",
            ],
            ["https://foss.heptapod.net/user1", "https://foss.heptapod.net"],
            id="known-suffix-hg",
        ),
        pytest.param(
            "svn+ssh://example.com/svnroot/users/user1/test1.svn",
            [
                "svn+ssh://example.com/svnroot/users/user1/test1.svn",
                "svn+ssh://example.com/svnroot/users/user1/test1",
            ],
            [
                "svn+ssh://example.com/svnroot/users/user1",
                "svn+ssh://example.com/svnroot/users",
                "svn+ssh://example.com/svnroot",
                "svn+ssh://example.com",
            ],
            id="known-suffix-svn",
        ),
        pytest.param(
            "https://github.com/user1/test1",
            [
                "https://github.com/user1/test1",
            ],
            ["https://github.com/user1", "https://github.com"],
            id="bare",
        ),
        pytest.param(
            "https://github.com/user1/test1/",
            [
                "https://github.com/user1/test1",
                "https://github.com/user1/test1/",
            ],
            ["https://github.com/user1", "https://github.com"],
            id="slash",
        ),
        pytest.param(
            "https://github.com/user1/test1.git/",
            [
                "https://github.com/user1/test1.git",
                "https://github.com/user1/test1.git/",
                "https://github.com/user1/test1",
            ],
            ["https://github.com/user1", "https://github.com"],
            id="slash-suffix",
        ),
    ],
)
def test_get_urls_to_check(url, exact, prefix):
    assert get_urls_to_check(url) == (exact, prefix)


def test_db_version(blocking_admin: BlockingAdmin):
    dbmodule, dbversion, dbflavor = get_database_info(blocking_admin.conn)
    assert dbmodule == "storage.proxies.blocking"
    assert dbversion == BlockingAdmin.current_version
    assert dbflavor is None


def test_create_find_request(blocking_admin: BlockingAdmin):
    created = blocking_admin.create_request(slug="foo", reason="bar")

    assert created.slug == "foo"
    assert created.reason == "bar"
    assert created.id is not None
    assert created.date is not None

    assert blocking_admin.find_request("foo") == created


def test_create_request_conflict(blocking_admin: BlockingAdmin):
    blocking_admin.create_request(slug="foo", reason="bar")

    with pytest.raises(DuplicateRequest) as exc_info:
        blocking_admin.create_request(slug="foo", reason="quux")

    assert exc_info.value.args == ("foo",)


def test_find_request_not_found(blocking_admin):
    slug = "notfound"
    assert blocking_admin.find_request(slug=slug) is None


def test_find_request_by_id(blocking_admin: BlockingAdmin):
    created = blocking_admin.create_request(slug="foo", reason="bar")

    assert blocking_admin.find_request_by_id(created.id) == created


NON_EXISTING_UUID: uuid.UUID = uuid.UUID("da785a27-acab-4a35-b82a-a5ae3714407c")


def test_find_request_by_id_not_found(blocking_admin: BlockingAdmin):
    assert blocking_admin.find_request_by_id(NON_EXISTING_UUID) is None


@pytest.fixture
def populated_blocking_admin(blocking_admin):
    pending_request1 = blocking_admin.create_request(slug="pending1", reason="one")
    blocking_admin.set_origins_state(
        request_id=pending_request1.id,
        new_state=BlockingState.DECISION_PENDING,
        urls=[StorageData.origin.url],
    )
    pending_request2 = blocking_admin.create_request(slug="pending2", reason="two")
    blocking_admin.set_origins_state(
        request_id=pending_request2.id,
        new_state=BlockingState.NON_BLOCKED,
        urls=[StorageData.origin.url],
    )
    blocking_admin.set_origins_state(
        request_id=pending_request2.id,
        new_state=BlockingState.DECISION_PENDING,
        urls=[StorageData.origin2.url],
    )
    blocking_admin.create_request(slug="cleared", reason="handled")
    # We add no blocks to this last one
    return blocking_admin


def test_get_requests_excluding_cleared_requests(populated_blocking_admin):
    assert [
        (request.slug, count)
        for request, count in populated_blocking_admin.get_requests(
            include_cleared_requests=False
        )
    ] == [("pending2", 2), ("pending1", 1)]


def test_get_requests_including_cleared_requests(populated_blocking_admin):
    assert [
        (request.slug, count)
        for request, count in populated_blocking_admin.get_requests(
            include_cleared_requests=True
        )
    ] == [("cleared", 0), ("pending2", 2), ("pending1", 1)]


def test_get_states_for_request(populated_blocking_admin):
    request = populated_blocking_admin.find_request("pending2")
    states = populated_blocking_admin.get_states_for_request(request.id)
    assert states == {
        StorageData.origin.url: BlockingState.NON_BLOCKED,
        StorageData.origin2.url: BlockingState.DECISION_PENDING,
    }


def test_get_states_for_request_not_found(blocking_admin: BlockingAdmin):
    with pytest.raises(RequestNotFound) as exc_info:
        blocking_admin.get_states_for_request(NON_EXISTING_UUID)
    assert exc_info.value.args == (NON_EXISTING_UUID,)


def test_find_blocking_states(populated_blocking_admin: BlockingAdmin):
    urls = [
        StorageData.origin.url,
        StorageData.origin2.url,
        # This one does not exist in the blocking db
        StorageData.origins[3].url,
    ]
    blocked_origins = populated_blocking_admin.find_blocking_states(urls)
    # The order in the output should be grouped by SWHID
    assert [
        (blocked.url_pattern, blocked.state, blocked.request_slug)
        for blocked in blocked_origins
    ] == [
        (StorageData.origin.url, BlockingState.NON_BLOCKED, "pending2"),
        (
            StorageData.origin.url,
            BlockingState.DECISION_PENDING,
            "pending1",
        ),
        (
            StorageData.origin2.url,
            BlockingState.DECISION_PENDING,
            "pending2",
        ),
    ]


def test_delete_blocking_states(populated_blocking_admin):
    request = populated_blocking_admin.find_request("pending2")
    populated_blocking_admin.delete_blocking_states(request.id)

    assert populated_blocking_admin.get_states_for_request(request.id) == {}


def test_delete_blocking_states_not_found(blocking_admin: BlockingAdmin):
    with pytest.raises(RequestNotFound) as exc_info:
        blocking_admin.delete_blocking_states(NON_EXISTING_UUID)
    assert exc_info.value.args == (NON_EXISTING_UUID,)


def test_record_history(blocking_admin: BlockingAdmin):
    messages = [f"message {i}" for i in range(3)]

    request = blocking_admin.create_request(slug="foo", reason="bar")

    for message in messages:
        msg = blocking_admin.record_history(request.id, message)
        assert msg.date is not None


def test_record_history_not_found(blocking_admin: BlockingAdmin):
    with pytest.raises(RequestNotFound) as exc_info:
        blocking_admin.record_history(NON_EXISTING_UUID, "kaboom")
    assert exc_info.value.args == (NON_EXISTING_UUID,)


def test_get_history(blocking_admin: BlockingAdmin):
    request = blocking_admin.create_request(slug="foo", reason="bar")
    blocking_admin.record_history(request.id, "one")
    blocking_admin.record_history(request.id, "two")
    blocking_admin.record_history(request.id, "three")

    history = blocking_admin.get_history(request.id)
    assert all(record.request == request.id for record in history)
    assert ["three", "two", "one"] == [record.message for record in history]


def test_get_history_not_found(blocking_admin: BlockingAdmin):
    with pytest.raises(RequestNotFound) as exc_info:
        blocking_admin.get_history(NON_EXISTING_UUID)
    assert exc_info.value.args == (NON_EXISTING_UUID,)


def test_swhid_lifecycle(blocking_admin: BlockingAdmin, blocking_query: BlockingQuery):
    # Create a request
    request = blocking_admin.create_request(slug="foo", reason="bar")

    all_origins = [origin.url for origin in StorageData.origins]
    blocked_origins = all_origins[::2]

    blocking_admin.set_origins_state(
        request_id=request.id,
        new_state=BlockingState.DECISION_PENDING,
        urls=blocked_origins,
    )

    expected = {
        url: BlockingStatus(state=BlockingState.DECISION_PENDING, request=request.id)
        for url in blocked_origins
    }

    assert blocking_query.origins_are_blocked(all_origins) == expected

    restricted = blocked_origins[0:2]

    blocking_admin.set_origins_state(
        request_id=request.id, new_state=BlockingState.BLOCKED, urls=restricted
    )

    for url in restricted:
        expected[url] = BlockingStatus(state=BlockingState.BLOCKED, request=request.id)

    assert blocking_query.origins_are_blocked(all_origins) == expected

    visible = blocked_origins[2:4]

    blocking_admin.set_origins_state(
        request_id=request.id, new_state=BlockingState.NON_BLOCKED, urls=visible
    )

    for url in visible:
        del expected[url]

    assert blocking_query.origins_are_blocked(all_origins) == expected

    for url in visible:
        expected[url] = BlockingStatus(
            state=BlockingState.NON_BLOCKED, request=request.id
        )

    assert (
        blocking_query.origins_are_blocked(all_origins, all_statuses=True) == expected
    )


def test_query_metrics(
    blocking_admin: BlockingAdmin, blocking_query: BlockingQuery, mocker
):
    increment = mocker.patch("swh.core.statsd.statsd.increment")

    # Create a request
    request = blocking_admin.create_request(slug="foo", reason="bar")

    all_origins = [origin.url for origin in StorageData.origins]
    blocked_origins = all_origins[::2]

    assert blocking_query.origins_are_blocked(all_origins) == {}
    assert increment.call_count == len(all_origins)
    increment.assert_has_calls(
        [call("swh_storage_blocking_queried_total", 1)] * len(all_origins)
    )

    increment.reset_mock()
    blocking_admin.set_origins_state(
        request_id=request.id,
        new_state=BlockingState.DECISION_PENDING,
        urls=blocked_origins,
    )

    assert len(blocking_query.origins_are_blocked(all_origins)) == len(blocked_origins)
    assert increment.call_args_list.count(
        call("swh_storage_blocking_queried_total", 1)
    ) == len(all_origins)
    assert increment.call_args_list.count(
        call("swh_storage_blocking_blocked_total", 1)
    ) == len(blocked_origins)



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import closing
import datetime as dt
from io import StringIO
import socket
import textwrap

from click.testing import CliRunner
import pytest

from swh.core.cli.db import db as swhdb
from swh.core.db.db_utils import get_database_info
from swh.core.tests.test_cli import assert_result
from swh.model.swhids import ExtendedSWHID, ValidationError

from ...proxies.masking.cli import (
    EditAborted,
    MaskedStateType,
    clear_request,
    edit_message,
    format_masked_state,
    history_cmd,
    list_requests,
    masking_cli_group,
    new_request,
    object_state,
    read_swhids,
    status_cmd,
    update_objects,
)
from ...proxies.masking.db import (
    MaskedObject,
    MaskedState,
    MaskingAdmin,
    MaskingRequest,
    MaskingRequestHistory,
    RequestNotFound,
)


def test_cli_db_create(postgresql):
    """Create a db then initializing it should be ok"""
    module_name = "storage:masking"
    db_params = postgresql.info
    dbname = f"{module_name}-db"
    conninfo = (
        f"postgresql://{db_params.user}@{db_params.host}:{db_params.port}/{dbname}"
    )

    # This creates the db and installs the necessary admin extensions
    result = CliRunner().invoke(swhdb, ["create", module_name, "--dbname", conninfo])
    assert_result(result)

    # This initializes the schema and data
    result = CliRunner().invoke(swhdb, ["init", module_name, "--dbname", conninfo])
    assert_result(result)

    dbmodule, dbversion, dbflavor = get_database_info(conninfo)
    assert dbmodule == "storage:masking"
    assert dbversion == MaskingAdmin.current_version
    assert dbflavor is None


@pytest.fixture
def masking_admin_config(masking_db_postgresql):
    return {
        "masking_admin": {"cls": "postgresql", "db": masking_db_postgresql.info.dsn}
    }


def test_masking_admin_not_defined():
    runner = CliRunner()
    result = runner.invoke(
        masking_cli_group,
        ["list-requests"],
        obj={"config": {}},
    )
    assert result.exit_code == 2
    assert "masking_admin" in result.output


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("localhost", 0))
        # We explicitly not set SO_REUSEADDR because we want the port
        # to stay free so we can get our connection refused.
        return s.getsockname()[1]


def test_masking_admin_unreachable():
    erroneous_postgresql_port = find_free_port()
    runner = CliRunner()
    result = runner.invoke(
        masking_cli_group,
        ["list-requests"],
        obj={
            "config": {
                "masking_admin": {
                    "db": (
                        "postgresql://localhost/postgres?"
                        f"port={erroneous_postgresql_port}"
                    )
                }
            }
        },
    )
    assert result.exit_code == 1
    assert "failed: Connection refused" in result.output


def test_edit_message_success(mocker):
    mocked_click_edit = mocker.patch(
        "swh.storage.proxies.masking.cli.click.edit", return_value="Sometimes I'm round"
    )
    assert (
        edit_message("Hello!", extra_lines=["Sometimes I'm alone", "Sometimes I'm not"])
        == "Sometimes I'm round"
    )
    mocked_click_edit.assert_called_once_with(
        textwrap.dedent(
            """

        # Hello!
        # Lines starting with “#” will be ignored. An empty message will abort the operation.
        #
        # Sometimes I'm alone
        # Sometimes I'm not"""
        )
    )


def test_edit_message_removes_comment(mocker):
    mocker.patch(
        "swh.storage.proxies.masking.cli.click.edit",
        return_value=textwrap.dedent(
            """\
        Hello!

        # This is a comment"""
        ),
    )
    assert edit_message("RANDOM PROMPT") == "Hello!"


def test_edit_message_empty_message_aborts(mocker):
    mocker.patch("swh.storage.proxies.masking.cli.click.edit", return_value="")
    with pytest.raises(EditAborted):
        edit_message("RANDOM PROMPT")


@pytest.fixture
def request01():
    return MaskingRequest(
        id="da785a27-7e59-4a35-b82a-a5ae3714407c",
        slug="request-01",
        date=dt.datetime.now(dt.timezone.utc),
        reason="one",
    )


@pytest.fixture
def request02():
    return MaskingRequest(
        id="4fd42e35-2b6c-4536-8447-bc213cd0118b",
        slug="request-02",
        date=dt.datetime.now(dt.timezone.utc),
        reason="two",
    )


@pytest.fixture
def masked_content(request01):
    return MaskedObject(
        request_slug=request01.slug,
        swhid=ExtendedSWHID.from_string(
            "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"
        ),
        state=MaskedState.RESTRICTED,
    )


@pytest.fixture
def masked_content2(request01):
    return MaskedObject(
        request_slug=request01.slug,
        swhid=ExtendedSWHID.from_string(
            "swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa"
        ),
        state=MaskedState.DECISION_PENDING,
    )


def masked_content3(request02):
    return MaskedObject(
        request_slug=request02.slug,
        swhid=ExtendedSWHID.from_string(
            "swh:1:cnt:c932c7649c6dfa4b82327d121215116909eb3bea"
        ),
        state=MaskedState.VISIBLE,
    )


@pytest.fixture
def history01(request01):
    return MaskingRequestHistory(
        request=request01.id,
        date=dt.datetime.now(dt.timezone.utc),
        message="record one",
    )


@pytest.fixture
def history02(request01):
    return MaskingRequestHistory(
        request=request01.id,
        date=dt.datetime.now(dt.timezone.utc),
        message="record two",
    )


@pytest.fixture
def mocked_masking_admin(
    mocker, request01, request02, masked_content, masked_content2, history01, history02
):
    def mock_find_request(slug):
        if slug == "request-01":
            return request01
        elif slug == "request-02":
            return request02
        else:
            return None

    def mock_find_request_by_id(id):
        if str(id) == "da785a27-7e59-4a35-b82a-a5ae3714407c":
            return request01
        return None

    def mock_get_states_for_request(request_id):
        if request_id == request01.id:
            return {
                masked_content.swhid: masked_content.state,
                masked_content2.swhid: masked_content2.state,
            }
        return RequestNotFound(request_id)

    def mock_get_history(request_id):
        if request_id == request01.id:
            return [history02, history01]
        return RequestNotFound(request_id)

    mock = mocker.Mock(spec=MaskingAdmin)
    mock.find_request.side_effect = mock_find_request
    mock.find_request_by_id.side_effect = mock_find_request_by_id
    mock.get_states_for_request.side_effect = mock_get_states_for_request
    mock.get_history.side_effect = mock_get_history
    mock.get_requests.return_value = [(request02, 2), (request01, 1)]
    return mock


def test_new_request(mocker, mocked_masking_admin, request01):
    mocked_masking_admin.create_request.return_value = request01
    mocker.patch("swh.storage.proxies.masking.cli.click.edit", return_value="one")
    runner = CliRunner()
    result = runner.invoke(
        new_request, ["request-01"], obj={"masking_admin": mocked_masking_admin}
    )
    assert_result(result)
    mocked_masking_admin.create_request.assert_called_once_with("request-01", "one")
    assert "da785a27-7e59-4a35-b82a-a5ae3714407c" in result.output
    assert "request-01" in result.output


def test_new_request_with_message(mocker, mocked_masking_admin, request01):
    mocked_masking_admin.create_request.return_value = request01
    mocked_click_edit = mocker.patch("swh.storage.proxies.masking.cli.click.edit")
    runner = CliRunner()
    result = runner.invoke(
        new_request,
        ["--message=one", "request-01"],
        obj={"masking_admin": mocked_masking_admin},
    )
    assert_result(result)
    mocked_click_edit.assert_not_called()
    mocked_masking_admin.create_request.assert_called_once_with("request-01", "one")
    assert "da785a27-7e59-4a35-b82a-a5ae3714407c" in result.output
    assert "request-01" in result.output


def test_new_request_empty_message_aborts(mocker, mocked_masking_admin):
    mocked_masking_admin.create_request.return_value = request01
    mocker.patch("swh.storage.proxies.masking.cli.click.edit", side_effect=EditAborted)
    runner = CliRunner()
    result = runner.invoke(
        new_request,
        ["request-01"],
        obj={"masking_admin": mocked_masking_admin},
    )
    assert result.exit_code == 1
    assert "Aborting" in result.output
    mocked_masking_admin.create_request.assert_not_called()


def test_list_requests_default(mocked_masking_admin):
    runner = CliRunner()
    result = runner.invoke(
        list_requests, [], obj={"masking_admin": mocked_masking_admin}
    )
    assert_result(result)
    mocked_masking_admin.get_requests.assert_called_once_with(
        include_cleared_requests=False
    )
    assert "request-01" in result.output
    assert "da785a27-7e59-4a35-b82a-a5ae3714407c" in result.output
    assert "request-02" in result.output
    assert "4fd42e35-2b6c-4536-8447-bc213cd0118b" in result.output


def test_list_requests_include_cleared_requests(mocked_masking_admin):
    runner = CliRunner()
    result = runner.invoke(
        list_requests,
        ["--include-cleared-requests"],
        obj={"masking_admin": mocked_masking_admin},
    )
    assert_result(result)
    mocked_masking_admin.get_requests.assert_called_once_with(
        include_cleared_requests=True
    )


SWHIDS_INPUT = """\
# content1
swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd

# content2
swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa
"""

SWHIDS_INPUT_PARSED = [
    ExtendedSWHID.from_string("swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
    ExtendedSWHID.from_string("swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa"),
]


def test_read_swhids():
    file = StringIO(SWHIDS_INPUT)
    assert read_swhids(file) == SWHIDS_INPUT_PARSED


def test_read_swhids_failure():
    file = StringIO("GARBAGE 💩")
    with pytest.raises(ValidationError):
        read_swhids(file)


def test_update_objects(mocker, mocked_masking_admin, request01):
    mocked_click_edit = mocker.patch(
        "swh.storage.proxies.masking.cli.click.edit", return_value="action made"
    )
    runner = CliRunner()
    result = runner.invoke(
        update_objects,
        ["request-01", "decision-pending"],
        input=SWHIDS_INPUT,
        obj={"masking_admin": mocked_masking_admin},
    )
    assert_result(result)
    assert "2 objects" in result.output
    assert "request-01" in result.output
    mocked_click_edit.assert_called_once()
    mocked_masking_admin.set_object_state.assert_called_once_with(
        request01.id, MaskedState.DECISION_PENDING, SWHIDS_INPUT_PARSED
    )
    mocked_masking_admin.record_history.assert_called_once_with(
        request01.id, "action made"
    )


def test_update_objects_with_message(mocker, mocked_masking_admin, request01):
    mocked_click_edit = mocker.patch(
        "swh.storage.proxies.masking.cli.click.edit", return_value="action made"
    )
    runner = CliRunner()
    result = runner.invoke(
        update_objects,
        ["request-01", "decision-pending", "--message=action made"],
        input=SWHIDS_INPUT,
        obj={"masking_admin": mocked_masking_admin},
    )
    assert_result(result)
    mocked_click_edit.assert_not_called()
    mocked_masking_admin.set_object_state.assert_called_once()
    mocked_masking_admin.record_history.assert_called_once_with(
        request01.id, "action made"
    )


def test_update_objects_empty_message_aborts(mocker, mocked_masking_admin):
    mocker.patch("swh.storage.proxies.masking.cli.click.edit", side_effect=EditAborted)
    runner = CliRunner()
    result = runner.invoke(
        update_objects,
        ["request-01", "decision-pending"],
        input=SWHIDS_INPUT,
        obj={"masking_admin": mocked_masking_admin},
    )
    assert result.exit_code == 1
    mocked_masking_admin.record_history.assert_not_called()


def test_update_objects_with_file(tmp_path, mocker, mocked_masking_admin, request01):
    mocker.patch(
        "swh.storage.proxies.masking.cli.click.edit", return_value="action made"
    )
    swhids_file = tmp_path / "swhids"
    swhids_file.write_text(SWHIDS_INPUT)
    runner = CliRunner()
    result = runner.invoke(
        update_objects,
        ["request-01", "decision-pending", f"--file={str(swhids_file)}"],
        obj={"masking_admin": mocked_masking_admin},
    )
    assert_result(result)
    assert "2 objects" in result.output
    mocked_masking_admin.set_object_state.assert_called_once_with(
        request01.id, MaskedState.DECISION_PENDING, SWHIDS_INPUT_PARSED
    )


def test_update_objects_unparseable_swhids(mocked_masking_admin):
    runner = CliRunner()
    result = runner.invoke(
        update_objects,
        ["request-01", "decision-pending", "--message=message"],
        input="GARBAGE 💩",
        obj={"masking_admin": mocked_masking_admin},
    )
    assert result.exit_code == 1
    assert "Invalid SWHID" in result.output
    mocked_masking_admin.set_object_state.assert_not_called()
    mocked_masking_admin.record_history.assert_not_called()


def test_update_objects_no_swhids(mocked_masking_admin):
    runner = CliRunner()
    result = runner.invoke(
        update_objects,
        ["request-01", "decision-pending", "--message=message"],
        input="",
        obj={"masking_admin": mocked_masking_admin},
    )
    assert result.exit_code == 1
    assert "No SWHIDs given!" in result.output
    mocked_masking_admin.set_object_state.assert_not_called()
    mocked_masking_admin.record_history.assert_not_called()


def test_status(mocked_masking_admin, request01):
    runner = CliRunner()
    result = runner.invoke(
        status_cmd, ["request-01"], obj={"masking_admin": mocked_masking_admin}
    )
    assert_result(result)
    mocked_masking_admin.get_states_for_request.assert_called_once_with(request01.id)
    assert result.output == textwrap.dedent(
        """\
        swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd restricted
        swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa decision-pending
        """
    )


def test_status_request_not_found(mocked_masking_admin):
    runner = CliRunner()
    result = runner.invoke(
        status_cmd, ["garbage"], obj={"masking_admin": mocked_masking_admin}
    )
    assert result.exit_code == 1
    assert "Error: Request “garbage” not found" in result.output


def test_status_request_via_uuid(mocked_masking_admin, request01):
    runner = CliRunner()
    result = runner.invoke(
        status_cmd,
        ["da785a27-7e59-4a35-b82a-a5ae3714407c"],
        obj={"masking_admin": mocked_masking_admin},
    )
    assert_result(result)
    mocked_masking_admin.get_states_for_request.assert_called_once_with(request01.id)


def test_history(mocked_masking_admin, request01):
    runner = CliRunner()
    result = runner.invoke(
        history_cmd, ["request-01"], obj={"masking_admin": mocked_masking_admin}
    )
    assert_result(result)
    mocked_masking_admin.get_history.assert_called_once_with(request01.id)
    assert (
        "History for request “request-01” (da785a27-7e59-4a35-b82a-a5ae3714407c)"
        in result.output
    )
    assert "record two" in result.output
    assert "record one" in result.output


def test_history_not_found(mocked_masking_admin):
    runner = CliRunner()
    result = runner.invoke(
        history_cmd, ["garbage"], obj={"masking_admin": mocked_masking_admin}
    )
    assert result.exit_code == 1
    assert "Error: Request “garbage” not found" in result.output


def test_object_state(mocked_masking_admin, masked_content, masked_content2):
    masked_content_in_request02 = MaskedObject(
        request_slug="request-02", swhid=masked_content.swhid, state=MaskedState.VISIBLE
    )
    mocked_masking_admin.find_masks.return_value = [
        masked_content_in_request02,
        masked_content,
        masked_content2,
    ]
    runner = CliRunner()
    result = runner.invoke(
        object_state,
        [
            "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd",
            "swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa",
            "swh:1:cnt:0000000000000000000000000000000000000000",
        ],
        obj={"masking_admin": mocked_masking_admin},
    )
    assert_result(result)
    assert result.output == textwrap.dedent(
        """\
        masked  swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd
                request-02: visible
                request-01: restricted
        masked  swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa
                request-01: decision-pending

        """
    )


def test_object_state_bad_swhid():
    runner = CliRunner()
    result = runner.invoke(
        object_state, ["GARBAGE"], obj={"masking_admin": mocked_masking_admin}
    )
    assert result.exit_code == 1
    assert "Unable to parse “GARBAGE” as a SWHID" in result.output


def test_clear_request(mocker, mocked_masking_admin, request01):
    mocked_click_edit = mocker.patch(
        "swh.storage.proxies.masking.cli.click.edit", return_value="request cleared"
    )
    runner = CliRunner()
    result = runner.invoke(
        clear_request,
        ["request-01"],
        obj={"masking_admin": mocked_masking_admin},
    )
    assert_result(result)
    assert "Masks cleared for request “request-01”." in result.output
    mocked_click_edit.assert_called_once()
    mocked_masking_admin.delete_masks.assert_called_once_with(request01.id)
    mocked_masking_admin.record_history.assert_called_once_with(
        request01.id, "request cleared"
    )


def test_clear_request_with_message(mocker, mocked_masking_admin, request01):
    mocked_click_edit = mocker.patch("swh.storage.proxies.masking.cli.click.edit")
    runner = CliRunner()
    result = runner.invoke(
        clear_request,
        ["--message=request cleared", "request-01"],
        obj={"masking_admin": mocked_masking_admin},
    )
    assert_result(result)
    assert "Masks cleared for request “request-01”." in result.output
    mocked_click_edit.assert_not_called()
    mocked_masking_admin.delete_masks.assert_called_once_with(request01.id)
    mocked_masking_admin.record_history.assert_called_once_with(
        request01.id, "request cleared"
    )


def test_clear_request_empty_message_aborts(mocker, mocked_masking_admin):
    mocker.patch("swh.storage.proxies.masking.cli.click.edit", side_effect=EditAborted)
    runner = CliRunner()
    result = runner.invoke(
        clear_request,
        ["request-01"],
        obj={"masking_admin": mocked_masking_admin},
    )
    assert result.exit_code == 1
    assert "Aborting" in result.output
    mocked_masking_admin.delete_masks.assert_not_called()
    mocked_masking_admin.record_history.assert_not_called()


def test_clear_request_not_found(mocked_masking_admin):
    runner = CliRunner()
    result = runner.invoke(
        clear_request, ["garbage"], obj={"masking_admin": mocked_masking_admin}
    )
    assert result.exit_code == 1
    assert "Error: Request “garbage” not found" in result.output
    mocked_masking_admin.delete_masks.assert_not_called()
    mocked_masking_admin.record_history.assert_not_called()


def test_masked_state_type():
    masked_states = set(MaskedStateType().choices)
    # check MaskedStateType choices match MaskedState enum
    assert masked_states == {format_masked_state(state) for state in MaskedState}
    # check MaskedStateType choices are referenced in command documentation
    for masked_state in masked_states:
        assert masked_state in update_objects.__doc__



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import defaultdict
from typing import Iterable, Tuple, TypeVar

from swh.model.swhids import (
    CoreSWHID,
    ExtendedObjectType,
    ExtendedSWHID,
    ObjectType,
    QualifiedSWHID,
)
from swh.storage.interface import StorageInterface

T = TypeVar("T", CoreSWHID, ExtendedSWHID, CoreSWHID | ExtendedSWHID)


def known_swhids(storage: StorageInterface, swhids: Iterable[T]) -> set[T]:
    """Query the storage to check if ``swhids`` exist.

    We group SWHIDs by type and then call the corresponding storage ``_missing`` method
    (directory_missing, snapshot_missing, etc.) and switch the results as we want to
    know what exists, not what's missing.

    As the storage does not use values from the qualifier (snapshot, etc.) we exclude
    ``QualifiedSWHID`` from this method because
    ``known_swhids(storage, swh:1:cnt:1234567890;visit=swh:1:snp:098654321)``
    could return that `swh:1:cnt:1234567890` exists even if `swh:1:snp:098654321`
    doesn't.

    Args:
        storage: a ``StorageInterface``
        swhid: a list of SWHIDs

    Raises:
        TypeError: received a ``QualifiedSWHID`` in the ``swhids`` list

    Returns:
        A set of SWHIDs found in the storage
    """

    grouped_swhids: dict[ObjectType | ExtendedObjectType, list[T]] = defaultdict(list)
    missing: set[Tuple[ObjectType | ExtendedObjectType, bytes]] = set()

    for swhid in swhids:
        if isinstance(swhid, QualifiedSWHID):
            raise TypeError(
                f"This method can't properly handle QualifiedSWHID like {swhid} "
                "but only CoreSWHID or ExtendedSWHID"
            )
        grouped_swhids[swhid.object_type].append(swhid)

    for object_type, objects in grouped_swhids.items():
        if object_type == ObjectType.CONTENT:
            storage_missing_method = storage.content_missing_per_sha1_git
        else:
            storage_missing_method = getattr(
                storage, f"{object_type.name.lower()}_missing"
            )

        missing |= set(
            (object_type, object_id)
            for object_id in storage_missing_method(
                [swhid.object_id for swhid in objects]
            )
        )
    return {
        swhid for swhid in swhids if (swhid.object_type, swhid.object_id) not in missing
    }


def swhid_is_known(storage: StorageInterface, swhid: CoreSWHID | ExtendedSWHID) -> bool:
    """Query the storage to check if ``swhid`` exists.

    A wrapper for :meth:`known_swhids` but for a single SWHID.

    Args:
        storage: a ``StorageInterface``
        swhid: a SWHID

    Returns:
        :const:`True` if ``swhid`` exists in the storage
    """
    return swhid in known_swhids(storage, [swhid])



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
import pytest

from swh.model.model import BaseModel, Content
from swh.storage.tests.storage_data import DEFAULT_ALGORITHMS, StorageData


def test_storage_data():
    data = StorageData()

    for attribute_key in [
        "contents",
        "skipped_contents",
        "directories",
        "revisions",
        "releases",
        "snapshots",
        "origins",
        "origin_visits",
        "fetchers",
        "authorities",
        "origin_metadata",
        "content_metadata",
        "extids",
    ]:
        for obj in getattr(data, attribute_key):
            assert isinstance(obj, BaseModel)


def test_storage_data_colliding_contents():
    data = StorageData()
    for h in DEFAULT_ALGORITHMS:
        hashes = set()
        values = set()
        for content in data.colliding_contents[h]:
            assert isinstance(content, Content)
            values.add(content.data)
            hashes.add(getattr(content, h))

        assert len(hashes) == 1
        assert len(values) == len(data.colliding_contents[h])


@pytest.mark.parametrize(
    "collection",
    ("directories", "git_revisions", "hg_revisions", "releases", "snapshots"),
)
def test_storage_data_hash(collection):
    data = StorageData()

    for obj in getattr(data, collection):
        assert (
            obj.compute_hash() == obj.id
        ), f"{obj.compute_hash().hex()} != {obj.id.hex()}"



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import hashlib

import pytest

from swh.model.swhids import CoreSWHID, ObjectType, QualifiedSWHID
from swh.storage.algos.swhid import known_swhids, swhid_is_known

from ..storage_data import StorageData


def test_known_swhids(swh_storage):
    swh_storage.snapshot_add([StorageData.snapshot])
    swh_storage.revision_add([StorageData.revision])
    swh_storage.release_add([StorageData.release])
    swh_storage.directory_add([StorageData.directory])
    swh_storage.content_add([StorageData.content])

    missing_snapshot_swhid = CoreSWHID.from_string(
        f"swh:1:snp:{hashlib.sha1(b'test snapshot').hexdigest()}"
    )
    missing_revision_swhid = CoreSWHID.from_string(
        f"swh:1:rev:{hashlib.sha1(b'test revision').hexdigest()}"
    )
    missing_release_swhid = CoreSWHID.from_string(
        f"swh:1:rel:{hashlib.sha1(b'test release').hexdigest()}"
    )
    missing_directory_swhid = CoreSWHID.from_string(
        f"swh:1:dir:{hashlib.sha1(b'test directory').hexdigest()}"
    )
    missing_content_swhid = CoreSWHID.from_string(
        f"swh:1:cnt:{hashlib.sha1(b'test content').hexdigest()}"
    )
    results = known_swhids(
        swh_storage,
        [
            StorageData.snapshot.swhid(),
            StorageData.revision.swhid(),
            StorageData.release.swhid(),
            StorageData.directory.swhid(),
            StorageData.content.swhid(),
            missing_snapshot_swhid,
            missing_revision_swhid,
            missing_release_swhid,
            missing_directory_swhid,
            missing_content_swhid,
        ],
    )

    assert len(results) == 5

    assert StorageData.snapshot.swhid() in results
    assert missing_snapshot_swhid not in results
    assert StorageData.revision.swhid() in results
    assert missing_revision_swhid not in results
    assert StorageData.release.swhid() in results
    assert missing_release_swhid not in results
    assert StorageData.directory.swhid() in results
    assert missing_directory_swhid not in results
    assert StorageData.content.swhid() in results
    assert missing_content_swhid not in results


def test_swhid_is_known(swh_storage):
    swh_storage.snapshot_add([StorageData.snapshot])
    missing_snapshot_swhid = CoreSWHID.from_string(
        f"swh:1:snp:{hashlib.sha1(b'test snapshot').hexdigest()}"
    )
    assert swhid_is_known(swh_storage, StorageData.snapshot.swhid())
    assert not swhid_is_known(swh_storage, missing_snapshot_swhid)


def test_does_not_handle_qualified_swhid(swh_storage):
    swhid = f"swh:1:snp:{hashlib.sha1(b'test qualified').hexdigest()}"
    qualified_swhid = QualifiedSWHID.from_string(swhid)
    with pytest.raises(TypeError) as exc:
        known_swhids(swh_storage, [qualified_swhid])
    assert swhid in str(exc.value)

    with pytest.raises(TypeError) as exc:
        swhid_is_known(swh_storage, qualified_swhid)
    assert swhid in str(exc.value)


def test_known_swhids_same_hash(swh_storage):
    swh_storage.content_add([StorageData.content])

    content_swhid = CoreSWHID(
        object_type=ObjectType.CONTENT, object_id=StorageData.content.sha1_git
    )

    directory_swhid = CoreSWHID(
        object_type=ObjectType.DIRECTORY, object_id=StorageData.content.sha1_git
    )

    assert known_swhids(swh_storage, [content_swhid, directory_swhid]) == {
        content_swhid
    }



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

import attr
import pytest

from swh.model.model import Directory, DirectoryEntry
from swh.storage.algos.directory import (
    directory_get,
    directory_get_many,
    directory_get_many_with_possibly_duplicated_entries,
)

from ..storage_data import StorageData


@pytest.mark.parametrize(
    "directory_id",
    [
        pytest.param(d.id, id=f"swh:dir:1:{d.id.hex()}")
        for i, d in enumerate(StorageData.directories)
    ],
)
def test_directory_small(swh_storage, directory_id):
    swh_storage.directory_add(StorageData.directories)

    (expected_directory,) = [d for d in StorageData.directories if d.id == directory_id]
    returned_directory = directory_get(swh_storage, directory_id)
    assert returned_directory.id == expected_directory.id
    assert set(returned_directory.entries) == set(expected_directory.entries)
    assert returned_directory.raw_manifest == expected_directory.raw_manifest


def test_directory_missing(swh_storage):
    swh_storage.directory_add(StorageData.directories)

    assert directory_get(swh_storage, b"\x42" * 20) is None


def test_directory_large(swh_storage):
    expected_directory = Directory(
        entries=tuple(
            DirectoryEntry(
                name=f"entry{i:04}".encode(),
                type="file",
                target=b"\x00" * 20,
                perms=0o000664,
            )
            for i in range(10)
        )
    )

    swh_storage.directory_add([expected_directory])

    returned_directory = directory_get(swh_storage, expected_directory.id)

    assert returned_directory.id == expected_directory.id
    assert set(returned_directory.entries) == set(expected_directory.entries)
    assert returned_directory.raw_manifest == expected_directory.raw_manifest


@pytest.mark.parametrize("with_possibly_duplicated_entries", [True, False])
def test_directories_small(swh_storage, with_possibly_duplicated_entries):
    swh_storage.directory_add(StorageData.directories)

    directory_ids = [d.id for d in StorageData.directories]

    if with_possibly_duplicated_entries:
        ret = list(
            directory_get_many_with_possibly_duplicated_entries(
                swh_storage, directory_ids
            )
        )
        assert not any(is_corrupt for (is_corrupt, _) in ret)
        returned_directories = [d for (_, d) in ret]
    else:
        returned_directories = list(directory_get_many(swh_storage, directory_ids))

    assert sorted(d.id for d in returned_directories) == sorted(directory_ids)
    for returned_directory in returned_directories:
        (expected_directory,) = [
            d for d in StorageData.directories if d.id == returned_directory.id
        ]
        assert set(returned_directory.entries) == set(expected_directory.entries)
        assert returned_directory.raw_manifest == expected_directory.raw_manifest


@pytest.mark.parametrize("with_possibly_duplicated_entries", [True, False])
def test_directories_missing(swh_storage, with_possibly_duplicated_entries):
    swh_storage.directory_add(StorageData.directories)

    missing_ids = [b"\x42" * 20, b"\x24" * 20]
    directory_ids = [d.id for d in StorageData.directories]
    directory_ids += missing_ids
    random.shuffle(directory_ids)

    if with_possibly_duplicated_entries:
        ret = list(
            directory_get_many_with_possibly_duplicated_entries(
                swh_storage, directory_ids
            )
        )
        assert not any(
            False if x is None else x[0] for x in ret
        ), "Unexpectedly marked corrupt"
        returned_directories = [None if x is None else x[1] for x in ret]
    else:
        returned_directories = list(directory_get_many(swh_storage, directory_ids))

    assert [d and d.id for d in returned_directories] == [
        None if id_ in missing_ids else id_ for id_ in directory_ids
    ]


def test_directories_large(swh_storage):
    expected_directory = Directory(
        entries=tuple(
            DirectoryEntry(
                name=f"entry{i:04}".encode(),
                type="file",
                target=b"\x00" * 20,
                perms=0o000664,
            )
            for i in range(10)
        )
    )

    swh_storage.directory_add([expected_directory])

    returned_directories = list(
        directory_get_many(swh_storage, [expected_directory.id, b"\x42" * 20])
    )

    (returned_directory, none) = returned_directories
    assert none is None

    assert returned_directory.id == expected_directory.id
    assert set(returned_directory.entries) == set(expected_directory.entries)
    assert returned_directory.raw_manifest == expected_directory.raw_manifest


@pytest.mark.parametrize("allow_duplicated_entries", [False, True])
def test_directories_duplicate_entries(swh_storage, allow_duplicated_entries):
    run_validators = attr.get_run_validators()
    attr.set_run_validators(False)
    try:
        invalid_directory = Directory(
            entries=(
                DirectoryEntry(name=b"foo", type="dir", target=b"\x01" * 20, perms=1),
                DirectoryEntry(name=b"foo", type="file", target=b"\x00" * 20, perms=0),
            )
        )
    finally:
        attr.set_run_validators(run_validators)
    swh_storage.directory_add([invalid_directory])

    deduplicated_directory = Directory(
        id=invalid_directory.id,
        entries=(
            DirectoryEntry(name=b"foo", type="dir", target=b"\x01" * 20, perms=1),
            DirectoryEntry(
                name=b"foo_0000000000", type="file", target=b"\x00" * 20, perms=0
            ),
        ),
        raw_manifest=(
            # fmt: off
            b"tree 52\x00"
            + b"0 foo\x00" + b"\x00" * 20
            + b"1 foo\x00" + b"\x01" * 20
            # fmt: on
        ),
    )

    # Make sure we successfully inserted a corrupt directory, otherwise this test
    # is pointless
    db = swh_storage.get_db()
    with db.conn.cursor() as cur:
        cur.execute("select id, dir_entries, file_entries, raw_manifest from directory")
        (row,) = cur
        (id_, (dir_entry,), (file_entry,), raw_manifest) = row
        assert id_ == invalid_directory.id
        assert raw_manifest is None
        cur.execute("select id, name, target from directory_entry_dir")
        assert list(cur) == [(dir_entry, b"foo", b"\x01" * 20)]
        cur.execute("select id, name, target from directory_entry_file")
        assert list(cur) == [(file_entry, b"foo", b"\x00" * 20)]

    if allow_duplicated_entries:
        ((is_corrupt, returned_directory),) = list(
            directory_get_many_with_possibly_duplicated_entries(
                swh_storage, [invalid_directory.id]
            )
        )

        assert is_corrupt
        assert (
            returned_directory.id == invalid_directory.id == deduplicated_directory.id
        )
        assert set(returned_directory.entries) == set(deduplicated_directory.entries)
        assert returned_directory.raw_manifest == deduplicated_directory.raw_manifest
    else:
        with pytest.raises(ValueError):
            list(directory_get_many(swh_storage, [invalid_directory.id]))
